from .helpers import tmp_storage


def test_x_oca_event(ccoe_bundle, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [ccoe_bundle])

    store.extract('events', 'x-oca-event', 'q1', "[x-oca-event:kind = 'event']")
    print(store.columns('events'))
    assert 'process_ref.id' in store.columns('events')



import logging
import os
from collections import OrderedDict, defaultdict

import orjson

from firepit import raft
from firepit.exceptions import DuplicateTable


logger = logging.getLogger(__name__)


def _strip_prefix(s, prefix):
    return s[len(prefix):] if s.startswith(prefix) else s


class JsonWriter:
    """
    Writes STIX objects to JSON files, one file per type
    """

    def __init__(self, filedir):
        self.filenames = {}
        self.files = {}
        self.filedir = filedir
        self.props = {}

    def _get_fp(self, obj_type):
        if obj_type not in self.files:
            filename = os.path.join(self.filedir, obj_type) + '.json'
            self.files[obj_type] = open(filename, 'w')
            self.filenames[obj_type] = filename
        return self.files[obj_type]

    @staticmethod
    def infer_type(key, value):
        return 'N/A'

    def new_type(self, obj_type, schema):
        self.props[obj_type] = schema

    def new_property(self, obj_type, prop_name, prop_type):
        self.props[obj_type][prop_name] = prop_type

    def write_records(self, obj_type, records, schema, replace, query_id):
        if replace:
            raise Exception('"replace" not supported when writing JSON')
        fp = self._get_fp(obj_type)
        #buf = '\n'.join([str(orjson.dumps(rec), 'utf-8') for rec in records]) + '\n'
        #fp.write('{}\n'.format(buf))
        for record in records:
            obj = OrderedDict(zip(schema.keys(), record))
            buf = str(orjson.dumps(obj), 'utf-8')
            fp.write(buf)

    def types(self):
        return list(self.files.keys())

    def properties(self, obj_type):
        return self.props.get(obj_type, {})

    def __del__(self):
        for fp in self.files.values():
            if fp:
                fp.close()


class SqlWriter:
    """
    Writes STIX objects to a SQL DB, one table per type
    """

    def __init__(self, filedir, store, prefix='',
                 placeholder='?',
                 infer_type=None,
                 **kwargs):
        self.filedir = filedir
        self.store = store
        if prefix and not prefix.endswith('_'):
            self.prefix = prefix + '_'
        else:
            self.prefix = prefix
        self.placeholder = placeholder
        self.infer_type = infer_type
        self.schemas = defaultdict(OrderedDict)
        self.kwargs = kwargs

    def _execute(self, stmt, cursor=None):
        return self.store._execute(stmt, cursor)

    def _replace(self, cursor, tablename, obj, schema):
        colnames = schema.keys()
        valnames = ', '.join([f'"{x}"' for x in colnames])
        placeholders = ', '.join([self.placeholder] * len(obj))
        stmt = f'INSERT INTO "{tablename}" ({valnames}) VALUES ({placeholders}) ON CONFLICT (id) DO '
        valnames = [f'"{col}" = EXCLUDED."{col}"' for col in colnames if col != 'id']
        valnames = ', '.join(valnames)
        stmt += f'UPDATE SET {valnames};'
        tmp = [str(orjson.dumps(value), 'utf-8')
               if isinstance(value, list) else value for value in obj]
        values = tuple(tmp)
        logger.debug('_replace: "%s" values %s', stmt, values)
        cursor.execute(stmt, values)

    def new_type(self, obj_type, schema):
        tablename = f'{self.prefix}{obj_type}'
        self.store._create_table(tablename, schema)

    def new_property(self, obj_type, prop_name, prop_type):
        tablename = f'{self.prefix}{obj_type}'
        self.store._add_column(tablename, prop_name, prop_type)

    def write_records(self, obj_type, records, schema, replace, query_id):
        tablename = f'{self.prefix}{obj_type}'
        try:
            self.store.connection.commit()
            cursor = self.store.connection.cursor()
            cursor.execute('BEGIN')
            if replace:
                for obj in records:
                    self._replace(cursor, tablename, obj, schema)
            else:
                kwargs = {k: v for k, v in self.kwargs.items() if k != 'query_id'}
                self.store.upsert_many(cursor, tablename, records, query_id, schema, **kwargs)
            cursor.execute('COMMIT')
        finally:
            cursor.close()

    def types(self):
        tables = self.store.tables()
        return [_strip_prefix(table, self.prefix) for table in tables]

    def properties(self, obj_type):
        tablename = f'{self.prefix}{obj_type}'
        return self.store.schema(tablename)


class SplitWriter:
    """
    Writes STIX objects using `writer`.  This class will track schema
    changes and store `batchsize` objects in memory before passing to
    `writer`.

    """

    def __init__(self, writer, batchsize=10, extras=None, replace=False, query_id=None):
        self.schemas = {}
        self.writer = writer
        self.batchsize = batchsize
        self.records = defaultdict(list)
        self.extras = extras or {}
        self.replace = replace
        self.query_id = query_id
        self._load_schemas()

    def __del__(self):
        pass

    def _load_schema(self, obj_type):
        return {col['name']: col['type'] for col in self.writer.properties(obj_type)}

    def _load_schemas(self):
        for obj_type in self.writer.types():
            self.schemas[obj_type] = self._load_schema(obj_type)

    def write(self, obj):
        """Consume `obj` (actual writing to storage may be deferred)"""
        obj.update(self.extras)
        obj_type = obj['type']
        schema = self.schemas.get(obj_type)
        add_table = False
        add_col = False
        new_columns = {}
        if not schema:
            schema = {}
            add_table = True
        new_obj = {}
        for key, value in obj.items():
            if len(key) > 63:
                continue
            new_obj[key] = value
            if key not in schema:
                if not add_table:
                    add_col = True
                col_type = self.writer.infer_type(key, value)
                schema[key] = col_type
                new_columns[key] = col_type
        obj = new_obj
        if add_table:
            self.schemas[obj_type] = schema
            try:
                self.writer.new_type(obj_type, schema)
            except DuplicateTable:
                # Refresh schemas
                loaded_schema = self._load_schema(obj_type)
                new_columns = {}
                for key in set(schema) - set(loaded_schema):
                    new_columns[key] = schema[key]
                    add_col = True
        if add_col:
            for col, col_type in new_columns.items():
                for rec in self.records[obj_type]:
                    rec.append(None)
                self.writer.new_property(obj_type, col, col_type)
        self.records[obj_type].append([obj.get(col) for col in schema.keys()])
        if len(self.records[obj_type]) % self.batchsize == 0:
            self.writer.write_records(obj_type, self.records[obj_type], schema, self.replace, self.query_id)
            self.records[obj_type] = []

    def close(self):
        if self.batchsize > 1:
            for obj_type, recs in self.records.items():
                if recs:
                    # We've already added any necessary tables or columns
                    self.writer.write_records(obj_type, recs, self.schemas[obj_type], self.replace, self.query_id)


if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s %(levelname)s %(name)s: %(message)s', level=logging.DEBUG)
    import argparse
    parser = argparse.ArgumentParser('Split STIX bundles by object type')
    parser.add_argument('-d', '--directory', metavar='DIR', default='.')
    parser.add_argument('-f', '--format', metavar='FMT', default='json')
    parser.add_argument('-b', '--batchsize', metavar='N', default=100, type=int)
    parser.add_argument('-p', '--prefix', metavar='PREFIX', default='test_')
    parser.add_argument('-n', '--dbname', metavar='DBNAME', default='test.db')
    parser.add_argument('ops', metavar='OP,...')
    parser.add_argument('filename', metavar='FILENAME', nargs='+')
    args = parser.parse_args()

    if args.format == 'json':
        writer = JsonWriter(args.directory)
    elif args.format in ['sql', 'sqlite', 'sqlite3']:
        from sqlitestorage import SQLiteStorage
        store = SQLiteStorage(args.dbname)
        writer = store._get_writer()  # SqlWriter(args.directory, store, prefix=args.prefix)
    else:
        raise NotImplementedError(args.format)

    splitter = SplitWriter(writer, batchsize=args.batchsize)
    from pyinstrument import Profiler
    profiler = Profiler()
    profiler.start()
    for f in args.filename:
        for result in raft.transform(args.ops.split(','), f):
            splitter.write(result)
    profiler.stop()
    print(profiler.output_text(unicode=True, color=True))
    splitter.close()



import pytest

from firepit.props import auto_agg
from firepit.props import primary_prop


@pytest.mark.parametrize(
    'sco_type, expected', [
        ('directory', 'path'),
        ('file', 'name'),
        ('ipv4-addr', 'value'),
        ('ipv6-addr', 'value'),
        ('process', 'name'),
        ('url', 'value'),
        ('user-account', 'user_id'),
    ]
)
def test_primary_prop(sco_type, expected):
    assert primary_prop(sco_type) == expected


@pytest.mark.parametrize(
    'sco_type, prop, col_type, expected', [
        ('directory', 'path', 'TEXT', 'COUNT(DISTINCT "path") AS "unique_path"'),
        ('file', 'name', 'TEXT', 'COUNT(DISTINCT "name") AS "unique_name"'),
        ('file', 'first_observed', 'TEXT', 'MIN("first_observed") AS "first_observed"'),
        ('file', 'last_observed', 'TEXT', 'MAX("last_observed") AS "last_observed"'),
        ('file', 'number_observed', 'INTEGER', 'SUM("number_observed") AS "number_observed"'),
        ('file', 'hashes.MD5', 'TEXT', 'COUNT(DISTINCT "hashes.MD5") AS "unique_hashes.MD5"'),
        ('ipv4-addr', 'value', 'TEXT', 'COUNT(DISTINCT "value") AS "unique_value"'),
        ('ipv6-addr', 'value', 'TEXT', 'COUNT(DISTINCT "value") AS "unique_value"'),
        ('ipv6-addr', 'xf_risk', 'INTEGER', 'AVG("xf_risk") AS "mean_xf_risk"'),
        ('network-traffic', 'dst_bytes', 'INTEGER', 'AVG("dst_bytes") AS "mean_dst_bytes"'),
        ('network-traffic', 'dst_port', 'INTEGER', 'COUNT(DISTINCT "dst_port") AS "unique_dst_port"'),
        ('network-traffic', 'src_bytes', 'INTEGER', 'AVG("src_bytes") AS "mean_src_bytes"'),
        ('network-traffic', 'src_port', 'INTEGER', 'COUNT(DISTINCT "src_port") AS "unique_src_port"'),
        ('process', 'name', 'TEXT', 'COUNT(DISTINCT "name") AS "unique_name"'),
        ('process', 'pid', 'INTEGER', 'COUNT(DISTINCT "pid") AS "unique_pid"'),
        ('process', 'ppid', 'INTEGER', 'COUNT(DISTINCT "ppid") AS "unique_ppid"'),
        ('url', 'value', 'TEXT', 'COUNT(DISTINCT "value") AS "unique_value"'),
        ('url', 'id', 'TEXT', None),
        ('url', 'type', 'TEXT', None),
        ('url', 'x_contained_by_ref', 'TEXT', None),
        ('url', 'x_root', 'INTEGER', None),
        ('user-account', 'user_id', 'TEXT', 'COUNT(DISTINCT "user_id") AS "unique_user_id"'),
        ('ipv4-addr', 'xf_risk', 'bigint', 'AVG("xf_risk") AS "mean_xf_risk"'),
    ]
)
def test_auto_agg(sco_type, prop, col_type, expected):
    agg = auto_agg(sco_type, prop, col_type)
    assert agg == expected



#!/usr/bin/env python

"""
Raft: Streaming processing of STIX Observation SDOs (`observed-data`).
"""

from collections import OrderedDict
from collections import defaultdict

import ijson
import orjson
import ujson
import requests

INVERTED_REFS = {
    'x_child_of_ref': 'parent_ref',
    'x_resolves_to_of_ref': 'x_resolves_from_ref',
    'x_encapsulates_of_ref': 'encapsulated_by_ref',
}


class GeneratorIO:
    '''Convert a generator into a file-like object'''

    def __init__(self, gen):
        self.gen = gen
        self.buf = b''

    def read(self, n):
        result = b''
        try:
            while n > len(self.buf):
                self.buf += next(self.gen)
            result = self.buf[:n]
            self.buf = self.buf[n:]
        except StopIteration:
            result = self.buf
            self.buf = b''
        return result


def _get_objects(fp, types):
    try:
        for obj in ijson.items(fp, 'objects.item'):
            if not types or obj['type'] in types:
                yield obj
    except ijson.common.IncompleteJSONError:
        pass


def _yield_objects(bundle, types):
    if 'type' not in bundle or bundle['type'] != 'bundle':
        bundle = {}
    for obj in bundle.get('objects', []):
        if not types or obj.get('type') in types:
            yield obj


def get_objects(source, types=None):
    '''A generator function that yields STIX objects from source'''
    if isinstance(source, dict):
        yield from _yield_objects(source, types)
    elif source.startswith('http'):
        response = requests.get(source, stream=True)
        fp = GeneratorIO(response.iter_content(chunk_size=8192))
        yield from _get_objects(fp, types)
    elif hasattr(source, 'read'):
        yield from _get_objects(source, types)
    else:
        with open(source, 'r') as fp:
            bundle = orjson.loads(fp.read())
        yield from _yield_objects(bundle, types)


def _is_custom_obj(obj):
    return obj['type'].startswith('x-')


def _is_custom_prop(prop):
    return prop.startswith('x_')


def _set_id(idx, obs, oid):
    stype = obs['type']
    uid = oid.lstrip('observed-data')
    obs['id'] = f'{stype}--{uid}_{idx}'
    return uid


def preserve(obj):
    '''Stash the "raw" STIX JSON as a stirng attribute in the object itself'''
    if obj['type'] == 'observed-data':
        obj['x_stix'] = ujson.dumps(obj)
    return obj


def promote(obj):
    '''Given an Observation, "promote" contained SCOs to top level'''
    # TODO: also "promote"/synthesize SROs for references?
    observables = obj['objects']
    oid = obj['id']
    object_refs = []
    results = []
    if isinstance(observables, dict):
        for idx, obs in observables.items():
            obs['x_contained_by_ref'] = oid
            for prop in ['first_observed', 'last_observed', 'number_observed']:
                if prop in obj:
                    obs[prop] = obj[prop]
            uid = _set_id(idx, obs, oid)
            for prop, val in obs.items():
                if prop.endswith('_ref') or prop.endswith('_refs'):
                    if not isinstance(val, list):
                        val = [val]
                    for ref in val:
                        if ref in observables and ref != idx:  # Avoid bogus references
                            target = observables[ref]
                            obs[prop] = f'{target["type"]}--{uid}_{ref}'
            object_refs.append(obs['id'])
            results.append(obs)
    elif isinstance(observables, list):
        # Already ran through nest?
        for idx, orig_obs in enumerate(observables):
            obs = orig_obs.copy()
            obs['x_contained_by_ref'] = oid
            for prop in ['first_observed', 'last_observed', 'number_observed']:
                if prop in obj:
                    obs[prop] = obj[prop]
            object_refs.append(obs['id'])
            results.append(obs)
    del obj['objects']
    results.append(obj)
    return results


def makeid(obj):
    '''
    Deprecated: moved into `markroot`
    Add unique object IDs to SCOs inside an Observation SDO
    '''
    oid = obj['id']
    observables = obj.get('objects', {})
    for idx, obs in observables.items():
        _set_id(idx, obs, oid)
    return obj


def _resolve(obs_orig, observables):
    obs = {}
    is_custom_obj = _is_custom_obj(obs_orig)  #obs_orig['type'].startswith('x-')
    for prop, val in obs_orig.items():
        if ((prop.endswith('_ref') or prop.endswith('_refs')) and
            prop != 'child_refs'):
            id_only = False
            if is_custom_obj and not _is_custom_prop(prop):  # prop.startswith('x_'):
                id_only = True
            if not isinstance(val, list):
                if val in observables:
                    target = observables[val]
                    if id_only:
                        obs[prop] = {'id': target['id']}
                    elif not _is_custom_obj(target):
                        target = _resolve(target, observables)
                        obs[prop] = target
            else:
                obs[f'{prop}_count'] = len(val)
                for i, ref in enumerate(val):
                    if ref in observables:  # Avoid bogus references
                        target = observables[ref]
                        target = _resolve(target, observables)
                        obs[f'{prop}[{i}]'] = target
        else:
            obs[prop] = val
    return obs


# TODO: rename to "dereference"?  "resolve"?
def nest(obj):
    '''Resolve refs with deep copy'''
    object_refs = []
    observables = obj.get('objects', {})
    for _, obs_orig in observables.items():
        object_refs.append(_resolve(obs_orig, observables))
    del obj['objects']
    obj['objects'] = object_refs
    return obj


def json_normalize(d, prefix='', sep='.', flat_lists=True):
    r = OrderedDict()  # {}
    for k, v in d.items():
        if '-' in k:  # Weird STIX rule: single quotes around things like SHA-1
            if ':' in k:
                otype, _, path = k.rpartition(':')
                parts = path.split('.')
                key = f"{otype}:" + '.'.join([f"'{part}'" if '-' in part else part for part in parts])
            else:
                key = f"'{k}'"
        else:
            key = k
        if prefix:
            key = f'{prefix}{sep}{key}'
        if isinstance(v, dict):
            r.update(json_normalize(v, key, sep, flat_lists))
        elif flat_lists and isinstance(v, list):
            for i, val in enumerate(v):
                r[f'{key}[{i}]'] = val
        else:
            r[key] = v
    return r


def normalize(obj):
    """Normalize obj to a flat dict"""
    return json_normalize(obj, flat_lists=False)


def invert(obj):
    '''Invert reference lists so all refs are 1:1'''
    objects = obj['objects']

    for k, v in objects.items():
        reflists = []
        for attr, val in v.items():
            if attr.endswith('_refs') and isinstance(val, list):
                refname = 'x_' + attr.rstrip('_refs') + '_of_ref'
                refname = INVERTED_REFS.get(refname, refname)
                for ref in val:
                    if ref not in objects or ref == k:  # Detect bogus refs
                        continue
                    target = objects[ref]
                    target[refname] = k
                reflists.append(attr)

        for attr in reflists:
            # Record the count of the ref list we've inverted
            v['x_' + attr.rstrip('_refs') + '_count'] = len(v[attr])
            del v[attr]

    return obj


def _mark_tree(objs, k, reffed):
    reffed.add(k)
    for attr, val in objs[k].items():
        if attr.endswith('_ref'):
            if val not in objs or val == k:
                continue
            _mark_tree(objs, val, reffed)
        elif attr.endswith('_refs'):
            for ref in val:
                if ref not in objs or ref == k:
                    continue
                _mark_tree(objs, ref, reffed)


def markroot(obj):
    """
    If a SCO type appears multiple times in one observation, attempt
    to mark one as the "primary", or most significant, instance by
    setting an "x_root" attribute to 1

    The `makeid` transform was rolled into here as an optimization.

    """
    objs = obj['objects']
    reffed = set()
    oid = obj['id']

    # Keep track of the preference order of each reffed object, by type
    prefs = defaultdict(list)
    for idx, sco in objs.items():
        _set_id(idx, sco, oid)
        prefs[sco['type']].append(idx)
        for attr, val in sco.items():
            if attr.endswith('_ref'):
                if val not in objs or val == idx:
                    continue
                # If an object refs another object of the same type,
                # only mark the root (think process:parent_ref)
                if objs[idx]['type'] == objs[val]['type']:
                    _mark_tree(objs, val, reffed)
                elif (objs[val]['type'].endswith('-addr')):
                    if 'dst_' in attr:
                        # For src/dst pairs, consider the src as the root (so add dst to reffed)
                        reffed.add(val)
                    elif attr.endswith('src_ref'):
                        # Save ref as the "preferred" object for this type
                        prefs[objs[val]['type']].insert(0, val)
                elif val in reffed:
                    reffed.add(idx)
            elif attr.endswith('_refs'):
                for ref in val:
                    if ref not in objs or ref == idx:
                        continue
                    if objs[idx]['type'] == objs[ref]['type']:
                        reffed.add(ref)
    for k, v in objs.items():
        vtype = v['type']
        if k not in reffed:
            # Check if there's a more preferred object
            if vtype not in prefs:
                objs[k]['x_root'] = 1
            else:
                for i in prefs[vtype]:
                    if i in reffed:
                        continue
                    elif i == k:
                        objs[k]['x_root'] = 1
                    break
    return obj


def transform(ops, filename):
    for obj in get_objects(filename, ['identity', 'observed-data']):
        if obj['type'] == 'observed-data':
            inputs = [obj]
            for op in ops:  # making ops composable might be overkill (plus hard)
                results = []
                for i in inputs:
                    r = globals()[op](i)
                    if isinstance(r, list):
                        results.extend(r)
                    else:
                        results.append(r)
                inputs = results
        else:
            results = [obj]
        for result in results:
            yield result


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser('raft - streaming operations on STIX')
    parser.add_argument('op', metavar='OP')
    parser.add_argument('filename', metavar='FILENAME')
    args = parser.parse_args()
    for result in transform(args.op.split(','), args.filename):
        print(str(orjson.dumps(result), 'utf-8'))



import logging
import os
import re
from functools import lru_cache

import orjson
import psycopg2
import psycopg2.extras
import ujson

from firepit.exceptions import DuplicateTable
from firepit.exceptions import InvalidAttr
from firepit.exceptions import UnexpectedError
from firepit.exceptions import UnknownViewname
from firepit.splitter import SqlWriter
from firepit.sqlstorage import SqlStorage
from firepit.sqlstorage import infer_type
from firepit.sqlstorage import validate_name

logger = logging.getLogger(__name__)


def get_storage(url, session_id):
    dbname = url.path.lstrip('/')
    return PgStorage(dbname, url.geturl(), session_id)


def _infer_type(key, value):
    # PostgreSQL type specializations
    rtype = None
    if isinstance(value, bool):
        rtype = 'BOOLEAN'
    else:
        # Fall back to defaults
        rtype = infer_type(key, value)
    return rtype


# PostgreSQL defaults for COPY text format
SEP = '\t'
TEXT_ESCAPE_TABLE = str.maketrans({
    '\\': '\\\\',
    '\n': '\\n',
    '\r': '\\r',
    SEP: f'\\{SEP}'
})


@lru_cache(maxsize=256, typed=True)
def _text_encode(value):
    if value is None:
        return r'\N'
    elif not isinstance(value, str):
        return str(value)
    # MUST "escape" special chars
    return value.translate(TEXT_ESCAPE_TABLE)


class ListToTextIO:
    """
    Convert an iterable of lists into a file-like object with
    PostgreSQL TEXT formatting
    """

    def __init__(self, objs, cols, sep=SEP):
        self.it = iter(objs)
        self.cols = cols
        self.sep = sep
        self.buf = ''

    def read(self, n):
        result = ''
        try:
            while n > len(self.buf):
                obj = next(self.it)
                vals = [ujson.dumps(val) if isinstance(val, list)
                        else _text_encode(val) for val in obj]
                self.buf += self.sep.join(vals) + '\n'
            result = self.buf[:n]
            self.buf = self.buf[n:]
        except StopIteration:
            result = self.buf
            self.buf = ''
        return result


class TuplesToTextIO:
    """
    Convert an iterable of tuples into a file-like object
    """

    def __init__(self, objs, cols, sep=SEP):
        self.it = iter(objs)
        self.cols = cols
        self.sep = sep
        self.buf = ''

    def read(self, n):
        result = ''
        try:
            while n > len(self.buf):
                obj = next(self.it)
                self.buf += self.sep.join(obj)
                self.buf += '\n'
            result = self.buf[:n]
            self.buf = self.buf[n:]
        except StopIteration:
            result = self.buf
            self.buf = ''
        return result


class PgStorage(SqlStorage):
    def __init__(self, dbname, url, session_id=None):
        super().__init__()
        self.placeholder = '%s'
        self.text_min = 'LEAST'
        self.text_max = 'GREATEST'
        self.ifnull = 'COALESCE'
        self.dbname = dbname
        self.infer_type = _infer_type
        self.defer_index = False
        if not session_id:
            session_id = 'firepit'
        self.session_id = session_id
        options = f'options=--search-path%3D{session_id}'
        sep = '&' if '?' in url else '?'
        connstring = f'{url}{sep}{options}'
        self.connection = psycopg2.connect(
            connstring,
            cursor_factory=psycopg2.extras.RealDictCursor)

        self._create_firepit_common_schema()
        if session_id:
            try:
                self._execute(f'CREATE SCHEMA IF NOT EXISTS "{session_id}";')
                # how to check if schema exists
            except psycopg2.errors.UniqueViolation:
                self.connection.rollback()

        self._execute(f'SET search_path TO "{session_id}", firepit_common;')

        stmt = ("SELECT (EXISTS (SELECT *"
                " FROM INFORMATION_SCHEMA.TABLES"
                " WHERE TABLE_SCHEMA = %s"
                " AND  TABLE_NAME = '__queries'))")
        res = self._query(stmt, (session_id,)).fetchone()
        done = list(res.values())[0] if res else False
        if not done:
            self._setup()

        logger.debug("Connection to PostgreSQL DB %s successful", dbname)

    def _create_firepit_common_schema(self):
        try:
            stmt = ("SELECT  schema_name"
                    " FROM INFORMATION_SCHEMA.SCHEMATA"
                    " WHERE  schema_name = 'firepit_common'")
            res = self._query(stmt).fetchone()
            done = list(res.values())[0] if res else False
            if not done:
                self._execute('CREATE SCHEMA IF NOT EXISTS "firepit_common";')
                cursor = self._execute('BEGIN;')
                self._execute('''CREATE FUNCTION firepit_common.match(pattern TEXT, value TEXT)
                                RETURNS boolean AS $$
                                    SELECT regexp_match(value, pattern) IS NOT NULL;
                            $$ LANGUAGE SQL;''', cursor=cursor)
                self._execute('''CREATE FUNCTION firepit_common.in_subnet(addr TEXT, net TEXT)
                                RETURNS boolean AS $$
                                    SELECT addr::inet <<= net::inet;
                            $$ LANGUAGE SQL;''', cursor=cursor)
                cursor.close()
        except psycopg2.errors.DuplicateFunction:
            self.connection.rollback()

    def _setup(self):
        cursor = self._execute('BEGIN;')
        try:
            # Do DB initization from base class
            stmt = ('CREATE UNLOGGED TABLE IF NOT EXISTS "__symtable" '
                    '(name TEXT, type TEXT, appdata TEXT);')
            self._execute(stmt, cursor)
            stmt = ('CREATE UNLOGGED TABLE IF NOT EXISTS "__queries" '
                    '(sco_id TEXT, query_id TEXT);')
            self._execute(stmt, cursor)
            self.connection.commit()
            cursor.close()
        except (psycopg2.errors.DuplicateFunction, psycopg2.errors.UniqueViolation):
            # We probably already created all these, so ignore this
            self.connection.rollback()

    def _get_writer(self, **kwargs):
        """Get a DB inserter object"""
        self.defer_index = kwargs.get('defer_index', self.defer_index)
        filedir = os.path.dirname(self.dbname)
        return SqlWriter(
            filedir,
            self,
            placeholder=self.placeholder,
            infer_type=_infer_type,
            **kwargs
        )

    def __del__(self):
        if self.connection:
            logger.debug("Closing PostgreSQL DB connection")
            self.connection.close()

    def _query(self, query, values=None, cursor=None):
        """Private wrapper for logging SQL query"""
        logger.debug('Executing query: %s', query)
        if not cursor:
            cursor = self.connection.cursor()
        if not values:
            values = ()
        try:
            cursor.execute(query, values)
        except psycopg2.errors.UndefinedColumn as e:
            self.connection.rollback()
            raise InvalidAttr(str(e)) from e
        except psycopg2.errors.UndefinedTable as e:
            self.connection.rollback()
            raise UnknownViewname(str(e)) from e
        self.connection.commit()
        return cursor

    def _create_table(self, tablename, columns):
        # Same as base class, but disable WAL
        stmt = f'CREATE UNLOGGED TABLE "{tablename}" ('
        stmt += ','.join([f'"{colname}" {coltype}' for colname, coltype in columns.items()])
        stmt += ');'
        logger.debug('_create_table: "%s"', stmt)
        try:
            cursor = self._execute(stmt)
            if not self.defer_index and 'x_contained_by_ref' in columns:
                self._execute(f'CREATE INDEX "{tablename}_obs" ON "{tablename}" ("x_contained_by_ref");', cursor)
            self.connection.commit()
            cursor.close()
        except (psycopg2.errors.DuplicateTable,
                psycopg2.errors.DuplicateObject,
                psycopg2.errors.UniqueViolation) as e:
            self.connection.rollback()
            raise DuplicateTable(tablename) from e

    def _add_column(self, tablename, prop_name, prop_type):
        stmt = f'ALTER TABLE "{tablename}" ADD COLUMN "{prop_name}" {prop_type};'
        logger.debug('new_property: "%s"', stmt)
        try:
            cursor = self._execute(stmt)
            self.connection.commit()
            cursor.close()
        except psycopg2.errors.DuplicateColumn:
            self.connection.rollback()

    def _create_empty_view(self, viewname, cursor):
        cursor.execute(f'CREATE VIEW "{viewname}" AS SELECT NULL as type WHERE 1<>1;')

    def _create_view(self, viewname, select, sco_type, deps=None, cursor=None):
        """Overrides parent"""
        validate_name(viewname)
        if not cursor:
            cursor = self._execute('BEGIN;')
        is_new = True
        if not deps:
            deps = []
        elif viewname in deps:
            is_new = False
            # Get the query that makes up the current view
            slct = self._get_view_def(viewname)
            if self._is_sql_view(viewname, cursor):
                self._execute(f'DROP VIEW IF EXISTS "{viewname}"', cursor)
            else:
                self._execute(f'ALTER TABLE "{viewname}" RENAME TO "_{viewname}"', cursor)
                slct = slct.replace(viewname, f'_{viewname}')
            # Swap out the viewname for its definition
            select = re.sub(f'"{viewname}"', f'({slct}) AS tmp', select)
        try:
            self._execute(f'CREATE OR REPLACE VIEW "{viewname}" AS {select}', cursor)
        except psycopg2.errors.UndefinedTable:
            # Missing dep?
            self.connection.rollback()
            cursor = self._execute('BEGIN;')
            self._create_empty_view(viewname, cursor)
        except psycopg2.errors.InvalidTableDefinition:
            # Usually "cannot drop columns from view"
            #logger.error(e, exc_info=e)
            self.connection.rollback()
            cursor = self._execute('BEGIN;')
            self._execute(f'DROP VIEW IF EXISTS "{viewname}";', cursor)
            self._execute(f'CREATE VIEW "{viewname}" AS {select}', cursor)
            is_new = False
        except psycopg2.errors.SyntaxError as e:
            # We see this on SQL injection attempts
            raise UnexpectedError(e.args[0]) from e
        if is_new:
            self._new_name(cursor, viewname, sco_type)
        return cursor

    def _get_view_def(self, viewname):
        cursor = self._query("SELECT definition"
                             " FROM pg_views"
                             " WHERE schemaname = %s"
                             " AND viewname = %s", (self.session_id, viewname))
        viewdef = cursor.fetchone()
        if viewdef:
            stmt = viewdef['definition'].rstrip(';').replace('\n', ' ')

            # PostgreSQL will "expand" the original "*" to the columns
            # that existed at that time.  We need to get the star back, to
            # match SQLite3's behavior.
            otype = self.table_type(viewname)
            return re.sub(f"SELECT *(\"?{otype}\"?\\.\"?['A-Za-z0-9_\\.-]+\"?,? *)+ FROM",
                          "SELECT * FROM", stmt)

        # Must be a table
        return f'SELECT * FROM "{viewname}"'

    def _is_sql_view(self, name, cursor=None):
        cursor = self._query("SELECT definition"
                             " FROM pg_views"
                             " WHERE schemaname = %s"
                             " AND viewname = %s", (self.session_id, name))
        viewdef = cursor.fetchone()
        return viewdef is not None

    def tables(self):
        cursor = self._query("SELECT table_name"
                             " FROM information_schema.tables"
                             " WHERE table_schema = %s"
                             "   AND table_type != 'VIEW'", (self.session_id, ))
        rows = cursor.fetchall()
        return [i['table_name'] for i in rows
                if not i['table_name'].startswith('__')]

    def types(self):
        stmt = ("SELECT table_name FROM information_schema.tables"
                " WHERE table_schema = %s AND table_type != 'VIEW'"
                "  EXCEPT SELECT name as table_name FROM __symtable")
        cursor = self._query(stmt, (self.session_id, ))
        rows = cursor.fetchall()
        # Ignore names that start with 1 or 2 underscores
        return [i['table_name'] for i in rows
                if not i['table_name'].startswith('_')]

    def columns(self, viewname):
        validate_name(viewname)
        cursor = self._query("SELECT column_name"
                             " FROM information_schema.columns"
                             " WHERE table_schema = %s"
                             " AND table_name = %s", (self.session_id, viewname))
        rows = cursor.fetchall()
        return [i['column_name'] for i in rows]

    def schema(self, viewname):
        validate_name(viewname)
        cursor = self._query("SELECT column_name AS name, data_type AS type"
                             " FROM information_schema.columns"
                             " WHERE table_schema = %s"
                             " AND table_name = %s", (self.session_id, viewname))
        return cursor.fetchall()

    def delete(self):
        """Delete ALL data in this store"""
        cursor = self._execute('BEGIN;')
        self._execute(f'DROP SCHEMA "{self.session_id}" CASCADE;', cursor)
        self.connection.commit()
        cursor.close()

    def upsert_many(self, cursor, tablename, objs, query_id, schema, **kwargs):
        use_copy = kwargs.get('use_copy')
        if use_copy:
            self.upsert_copy(cursor, tablename, objs, query_id, schema)
        else:
            self.upsert_multirow(cursor, tablename, objs, query_id, schema)

    def upsert_multirow(self, cursor, tablename, objs, query_id, schema):
        colnames = list(schema.keys())
        quoted_colnames = [f'"{x}"' for x in colnames]
        valnames = ', '.join(quoted_colnames)

        placeholders = ', '.join([f"({', '.join([self.placeholder] * len(colnames))})"] * len(objs))
        stmt = f'INSERT INTO "{tablename}" ({valnames}) VALUES {placeholders}'
        if 'id' in colnames:
            idx = colnames.index('id')
            if tablename == 'identity':
                action = 'NOTHING'
            else:
                excluded = self._get_excluded(colnames, tablename)
                action = f'UPDATE SET {excluded}'
            stmt += f' ON CONFLICT (id) DO {action}'
        values = []
        query_values = []
        for obj in objs:
            if query_id:
                query_values.append(obj[idx])
                query_values.append(query_id)
            values.extend([str(orjson.dumps(value), 'utf-8') if isinstance(value, list) else value for value in obj])
        cursor.execute(stmt, values)

        if query_id:
            # Now add to query table as well
            placeholders = ', '.join([f'({self.placeholder}, {self.placeholder})'] * len(objs))
            stmt = (f'INSERT INTO "__queries" (sco_id, query_id)'
                    f' VALUES {placeholders}')
            cursor.execute(stmt, query_values)

    def upsert_copy(self, cursor, tablename, objs, query_id, schema):
        colnames = list(schema.keys())
        quoted_colnames = [f'"{x}"' for x in colnames]
        valnames = ', '.join(quoted_colnames)

        # Create a temp table that copies the structure of `tablename`
        cursor.execute(f'CREATE TEMP TABLE tmp AS SELECT * FROM "{tablename}" WHERE 1=2;')

        # Create a generator over `objs` that returns text formatted objects
        copy_stmt = f"COPY tmp({valnames}) FROM STDIN WITH DELIMITER '{SEP}'"
        cursor.copy_expert(copy_stmt, ListToTextIO(objs, colnames, sep=SEP))

        # Now SELECT from TEMP table to real table
        stmt = (f'INSERT INTO "{tablename}" ({valnames})'
                f' SELECT {valnames} FROM tmp')
        if 'id' in colnames:
            if tablename == 'identity':
                action = 'NOTHING'
            else:
                excluded = self._get_excluded(colnames, tablename)
                action = f'UPDATE SET {excluded}'
            stmt += f'  ON CONFLICT (id) DO {action}'
        cursor.execute(stmt)

        # Don't need the temp table anymore
        cursor.execute('DROP TABLE tmp')

        if query_id and 'id' in colnames:
            # Now add to query table as well
            idx = colnames.index('id')
            copy_stmt = f"COPY __queries(sco_id, query_id) FROM STDIN WITH DELIMITER '{SEP}'"
            qobjs = [(obj[idx], query_id) for obj in objs]
            cursor.copy_expert(copy_stmt, TuplesToTextIO(qobjs, ['sco_id', 'query_id'], sep=SEP))

    def finish(self):
        if self.defer_index:
            cursor = self._execute('BEGIN;')
            for tablename in self.tables():
                if 'x_contained_by_ref' in self.columns(tablename):
                    self._execute(f'CREATE INDEX "{tablename}_obs" ON "{tablename}" ("x_contained_by_ref");', cursor)
            self.connection.commit()
            cursor.close()



import ast
import csv
import ujson
import pytest

from collections import Counter
from decimal import Decimal

from firepit.exceptions import IncompatibleType
from firepit.exceptions import UnknownViewname
from firepit.query import Aggregation
from firepit.query import Group
from firepit.query import Limit
from firepit.query import Query
from firepit.query import Order
from firepit.query import Table

from .helpers import tmp_storage


def test_local(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', fake_bundle_file)
    assert 'url' in store.tables()
    assert 'url' in store.types()
    assert not store._is_sql_view('url')

    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    assert 'urls' in store.views()
    assert 'urls' not in store.types()
    assert store._is_sql_view('urls')
    urls = store.values('url:value', 'urls')
    print(urls)
    assert len(urls) == 14
    assert 'http://www8.example.com/page/176' in urls
    assert 'http://www27.example.com/page/64' not in urls

    store.delete()


def test_in_memory(fake_bundle_file, tmpdir):
    with open(fake_bundle_file, 'r') as fp:
        bundle = ujson.loads(fp.read())

    store = tmp_storage(tmpdir)
    store.cache('q1', bundle)

    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    urls = store.values('url:value', 'urls')
    print(urls)
    assert len(urls) == 14
    assert 'http://www8.example.com/page/176' in urls
    assert 'http://www27.example.com/page/64' not in urls

    store.delete()


def test_basic(fake_bundle_file, fake_csv_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])

    store.extract('urls', 'url', 'q1', "[ipv4-addr:value ISSUBSET '192.168.0.0/16']")
    urls = store.values('url:value', 'urls')
    print(urls)
    assert len(urls) == 31
    assert 'http://www27.example.com/page/64' in urls
    assert store.count('urls') == 31

    urls1 = store.lookup('urls', limit=5)
    assert len(urls1) == 5
    urls2 = store.lookup('urls', limit=5, offset=2, cols="value,number_observed")
    assert len(urls2) == 5
    assert len(urls2[1].keys()) == 2
    assert 'number_observed' in urls2[1].keys()

    store.assign('sorted', 'urls', op='sort', by='value')
    urls = store.values('url:value', 'sorted')
    print('sorted:', urls)
    assert len(urls) == 31
    assert urls[0] == 'http://www11.example.com/page/108'

    # Now try to change urls, even though sorted is defined using it
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    urls = store.values('url:value', 'urls')
    print('reused:', urls)
    assert len(urls) == 14
    sorted_urls = store.values('url:value', 'sorted')
    print('sorted:', sorted_urls)
    assert len(sorted_urls) == 14  # Also changes...weird

    store.extract('a_ips', 'ipv4-addr', 'q1', "[ipv4-addr:value LIKE '10.%']")
    a_ips = store.values('ipv4-addr:value', 'a_ips')
    print(a_ips)
    assert len(a_ips) == 100
    assert '10.0.0.141' in a_ips

    store.extract('a_ips', 'ipv4-addr', 'q1', "[ipv4-addr:value LIKE '10.%']")
    a_ips = store.values('ipv4-addr:value', 'a_ips')
    print(a_ips)
    assert len(a_ips) == 100
    assert '10.0.0.141' in a_ips

    store.extract('users', 'user-account', 'q1', "[ipv4-addr:value LIKE '10.%']")
    users = store.values('user-account:account_login', 'users')
    print(users)
    assert len(users) == 100
    counter = Counter(users)
    assert counter['henry'] == 2
    assert counter['isabel'] == 12
    by = 'user-account:account_login'
    store.assign('grouped_users', 'users', op='group', by=by)
    cols = store.columns('grouped_users')
    _, _, by = by.rpartition(':')
    assert f'unique_{by}' not in cols
    grouped_users = store.lookup('grouped_users')
    print(grouped_users)
    henry = next((item for item in grouped_users if item['account_login'] == 'henry'), None)
    assert henry
    assert henry['number_observed'] == 2
    isabel = next((item for item in grouped_users if item['account_login'] == 'isabel'), None)
    assert isabel
    assert isabel['number_observed'] == 12

    with open(fake_csv_file, newline='') as fp:
        reader = csv.DictReader(fp)
        def infer_type(value):
            try:
                return ast.literal_eval(value)
            except Exception:
                return value
        data = [{key: infer_type(val) for key, val in row.items()} for row in reader]
        res = store.load('test_procs', data)
        assert res == 'process'
    rows = store.lookup('test_procs')
    assert len(rows) == 5
    assert isinstance(rows[0]['pid'], int) or isinstance(rows[0]['pid'], Decimal)
    ids = [row['id'] for row in rows]
    assert 'process--41eb677f-0335-49da-98b8-375e22f8c94e_0' in ids
    assert 'process--0bb2e61f-8c88-415d-bb7a-bcffc991c38e_0' in ids
    #assert rows[1]['binary_ref.parent_directory_ref.path'] == 'C:\\Windows\\System32'
    #assert rows[2]['parent_ref.command_line'] == 'C:\\windows\\system32\\cmd.exe /c "reg delete HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run /v caldera /f"'

    ips = ['10.0.0.1', '10.0.0.2']
    res = store.load('test_ips', ips, sco_type='ipv4-addr')
    assert res == 'ipv4-addr'
    rows = store.lookup('test_ips')
    assert len(rows) == 2
    for row in rows:
        assert row['type'] == 'ipv4-addr'
        assert row['value'] in ips

    store.delete()
    store = tmp_storage(tmpdir)
    assert len(store.tables()) == 0


def test_join(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('local_ips', 'ipv4-addr', 'q1', "[ipv4-addr:value LIKE '%']")

    res = store.load('test_ips', [
        {
            'type': 'ipv4-addr',
            'value': '10.0.0.201',
            'risk': 'high',
        },
        {
            'type': 'ipv4-addr',
            'value': '10.0.0.214',
            'risk': 'high',
        }
    ])

    store.join('marked', 'local_ips', 'value', 'test_ips', 'value')
    rows = store.lookup('marked')
    assert 'value' in rows[0]
    assert 'risk' in rows[0]
    for row in rows:
        if row['value'] in ['10.0.0.201', '10.0.0.214']:
            assert row['risk'] == 'high'
        else:
            assert row['risk'] is None


@pytest.mark.parametrize(
    'sco_type, prop, op, value, expected, unexpected', [
        ('url', 'value', 'LIKE', '%example.com/page/1%', 'http://www26.example.com/page/176', 'http://www67.example.com/page/264'),
        ('url', 'value', 'MATCHES', '^.*example.com/page/1[0-9]*$', 'http://www26.example.com/page/176', 'http://www67.example.com/page/264'),
        ('ipv4-addr', 'value', 'ISSUBSET', '10.0.0.0/8', '10.0.0.141', '192.168.212.97'),
        ('ipv4-addr', 'value', '=', '10.0.0.141', '10.0.0.141', '192.168.212.97'),
        ('network-traffic', 'dst_port', '<=', 1024, 22, 3128),
        ('user-account', 'account_login', 'IN', ('alice', 'bob', 'carol'), 'bob', 'david'),
        ('network-traffic', 'dst_ref.value', 'ISSUBSET', '10.0.0.0/25', '10.0.0.73', '10.0.0.197'),
    ]
)
def test_ops(fake_bundle_file, tmpdir, sco_type, prop, op, value, expected, unexpected):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    if isinstance(value, str):
        value = f"'{value}'"
    store.extract('data', sco_type, 'q1', f"[{sco_type}:{prop} {op} {value}]")
    data = store.values(f"{sco_type}:{prop}", 'data')
    assert expected in data
    assert unexpected not in data

    # Try the negation when appropriate
    if op in ['IN', 'LIKE', 'MATCHES', 'ISSUBSET', 'ISSUPERSET']:
        store.extract('data', sco_type, 'q1', f"[{sco_type}:{prop} NOT {op} {value}]")
        data = store.values(f"{sco_type}:{prop}", 'data')
        assert unexpected in data
        assert expected not in data


def test_grouping(fake_bundle_file, fake_csv_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])

    store.extract('conns', 'network-traffic', 'q1', "[network-traffic:dst_port < 1024]")
    store.assign('conns', 'conns', op='group', by='src_ref.value')
    srcs = store.values('src_ref.value', 'conns')
    assert srcs

    groups = store.lookup('conns')
    assert groups
    assert 'unique_dst_port' in groups[0].keys()


def test_extract(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])

    store.extract('conns', 'network-traffic', 'q1', "[network-traffic:dst_port < 1024]")
    store.assign('conns', 'conns', op='group', by='src_ref.value')
    srcs = store.values('src_ref.value', 'conns')
    assert srcs

    groups = store.lookup('conns')
    assert groups
    assert 'unique_dst_port' in groups[0].keys()


def test_schema(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])

    schema = store.schema('url')
    print(schema)
    columns = [i['name'] for i in schema]
    assert 'id' in columns
    assert 'type' in columns
    assert 'value' in columns


def test_filter(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    store.filter('urls', 'url', 'urls', "[url:value = 'http://www20.example.com/page/19']")
    urls = store.values('url:value', 'urls')
    assert len(urls) == 1
    assert 'http://www20.example.com/page/19' == urls[0]
    views = store.views()
    assert len(views) == 1
    assert views[0] == 'urls'


def test_filter2(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('ssh_conns', 'network-traffic', 'q1', "[network-traffic:dst_port = 22]")
    store.filter('ssh_ips', 'ipv4-addr', 'ssh_conns', "[network-traffic:dst_port = 22]")
    ssh_conns = store.lookup('ssh_conns')
    assert len(ssh_conns) == 29
    ssh_ips = store.lookup('ssh_ips')
    assert len(ssh_ips) == 29 # BUG?: * 2
    views = store.views()
    assert len(views) == 2


def test_reassign(fake_bundle_file, fake_csv_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])

    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    urls = store.lookup('urls')
    assert len(urls) == 14
    #print(ujson.dumps(urls, indent=4))

    # Simulate running some analytics to enrich these
    for url in urls:
        url['x_enrich'] = 1

    # Now reload into the same var
    store.reassign('urls', urls)
    rows = store.lookup('urls')
    print(ujson.dumps(rows, indent=4))
    assert len(rows) == len(urls)
    assert rows[0]['x_enrich'] == 1

    # Make sure original var length isn't modified
    urls = store.lookup('urls')
    assert len(urls) == 14

    # Original var's objects should have been updated
    assert urls[0]['x_enrich'] == 1


def test_reassign_after_grouping(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])

    store.extract('conns', 'network-traffic', 'q1', "[network-traffic:dst_port < 1024]")
    assert 'conns' not in store.tables()
    assert 'conns' not in store.types()
    assert 'conns' in store.views()
    store.assign('grouped_conns', 'conns', op='group', by='src_ref.value')
    assert 'grouped_conns' in store.views()
    grouped_conns = store.lookup('grouped_conns')

    # Simulate running some analytics to enrich these
    for grp in grouped_conns:
        grp['x_enrich'] = 1

    # Now reload into the same var
    store.reassign('grouped_conns', grouped_conns)
    rows = store.lookup('grouped_conns')
    #print(ujson.dumps(rows, indent=4))
    assert len(rows) == len(grouped_conns)
    assert rows[0]['x_enrich'] == 1

    # Now it's a table!!!
    assert 'grouped_conns' in store.tables()
    assert 'grouped_conns' in store.views()
    assert 'grouped_conns' not in store.types()

    # Can we still work with it?
    store.assign('x_conns', 'grouped_conns', op='sort', by='src_ref.value')
    rows = store.lookup('x_conns')
    #print(ujson.dumps(rows, indent=4))
    assert len(rows) == len(grouped_conns)

    # Can we reassign to that name?
    store.assign('grouped_conns', 'grouped_conns', op='sort', by='src_ref.value')
    print('PC: tables:', store.tables())
    print('PC: types:', store.types())
    print('PC: views:', store.views())
    assert 'grouped_conns' not in store.tables()  # Now it's a SQL view again!
    assert 'grouped_conns' not in store.types()
    assert 'grouped_conns' in store.views()
    rows = store.lookup('grouped_conns')
    print(ujson.dumps(rows, indent=4))
    assert len(rows) == len(grouped_conns)
    

def test_appdata(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('ssh_conns', 'network-traffic', 'q1', "[network-traffic:dst_port = 22]")
    data = {'foo': 99}
    store.set_appdata('ssh_conns', ujson.dumps(data))
    result = ujson.loads(store.get_appdata('ssh_conns'))
    assert data['foo'] == result['foo']
    assert len(result) == len(data)

    store2 = tmp_storage(tmpdir, clear=False)
    result = ujson.loads(store2.get_appdata('ssh_conns'))
    assert data['foo'] == result['foo']
    assert len(result) == len(data)


def test_viewdata(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('ssh_conns', 'network-traffic', 'q1', "[network-traffic:dst_port = 22]")
    ssh_data = {'foo': 99}
    store.set_appdata('ssh_conns', ujson.dumps(ssh_data))
    store.extract('dns_conns', 'network-traffic', 'q1', "[network-traffic:dst_port = 53]")
    dns_data = {'bar': 98}
    store.set_appdata('dns_conns', ujson.dumps(dns_data))

    results = store.get_view_data(['ssh_conns', 'dns_conns'])
    assert len(results) == 2
    for result in results:
        if result['name'] == 'ssh_conns':
            assert ssh_data == ujson.loads(result['appdata'])
        else:
            assert dns_data == ujson.loads(result['appdata'])


def test_duplicate(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)

    # Query once
    store.cache('q1', [fake_bundle_file])
    store.extract('urls1', 'url', 'q1', "[url:value LIKE '%page/1%']")
    urls1 = store.values('url:value', 'urls1')

    # Now query again - not reasonable, but simulates getting duplicate IDs from different sources
    store.cache('q2', [fake_bundle_file])
    store.extract('urls2', 'url', 'q2', "[url:value LIKE '%page/1%']")
    urls2 = store.values('url:value', 'urls2')

    assert len(urls1) == len(urls2)


def test_sort_same_name(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('urls', 'url', 'q1', "[ipv4-addr:value ISSUBSET '192.168.0.0/16']")
    urls1 = store.values('url:value', 'urls')
    print(urls1)
    assert len(urls1) == 31
    store.assign('urls', 'urls', op='sort', by='value')
    urls2 = store.values('url:value', 'urls')
    print(urls2)
    assert len(urls2) == 31
    assert set(urls1) == set(urls2)


def test_merge(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)

    store.cache('test-bundle', [fake_bundle_file])
    all_urls = set(store.values('url:value', 'url'))

    store.extract('urls1', 'url', 'test-bundle', "[url:value LIKE '%page/1%']")
    urls1 = set(store.values('url:value', 'urls1'))

    store.extract('urls2', 'url', 'test-bundle', "[url:value NOT LIKE '%page/1%']")
    urls2 = set(store.values('url:value', 'urls2'))

    assert urls1 | urls2 == all_urls

    store.merge('merged', ['urls1', 'urls2'])
    merged = set(store.values('url:value', 'merged'))
    assert merged == all_urls


def test_change_type(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])

    # Create a var `foo` of type url
    store.extract('foo', 'url', 'q1', "[url:value LIKE '%page/1%']")
    urls = store.values('url:value', 'foo')
    print(urls)
    assert len(urls) == 14

    # Create a var `sorted_foo` of type url that depends on `foo`
    store.assign('sorted_foo', 'foo', op='sort', by='value')

    store.extract('foo', 'ipv4-addr', 'q1', "[ipv4-addr:value ISSUBSET '192.168.0.0/16']")


def test_remove(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', fake_bundle_file)

    store.extract('urls1', 'url', 'q1', "[url:value LIKE '%page/1%']")
    urls1 = store.lookup('urls1')
    assert len(urls1) == 14

    store.extract('urls2', 'url', 'q1', "[url:value LIKE '%page/2%']")
    urls2 = store.lookup('urls2')
    assert len(urls2)

    store.remove_view('urls1')
    with pytest.raises(UnknownViewname):
        store.lookup('urls1')

    urls2 = store.lookup('urls2')
    assert len(urls2)


def test_rename(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', fake_bundle_file)

    store.extract('urls1', 'url', 'q1', "[url:value LIKE '%page/1%']")
    urls1 = store.lookup('urls1')
    assert len(urls1) == 14

    store.rename_view('urls1', 'urls2')
    with pytest.raises(UnknownViewname):
        store.lookup('urls1')

    urls2 = store.lookup('urls2')
    assert len(urls2) == 14


@pytest.mark.parametrize(
    'names', [
        (['urls1']),
        (['urls2']),
        (['urls1', 'urls2']),
    ]
)
def test_remove_after_merge(fake_bundle_file, tmpdir, names):
    store = tmp_storage(tmpdir)

    store.cache('test-bundle', [fake_bundle_file])
    all_urls = set(store.values('url:value', 'url'))

    store.extract('urls1', 'url', 'test-bundle', "[url:value LIKE '%page/1%']")
    urls1 = set(store.values('url:value', 'urls1'))

    store.extract('urls2', 'url', 'test-bundle', "[url:value NOT LIKE '%page/1%']")
    urls2 = set(store.values('url:value', 'urls2'))

    assert urls1 | urls2 == all_urls

    store.merge('merged', ['urls1', 'urls2'])

    # Remove the views we merged
    for name in names:
        store.remove_view(name)

    merged = set(store.values('url:value', 'merged'))
    assert merged == all_urls


def test_port_zero(fake_bundle_file_2, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file_2])

    store.extract('conns', 'network-traffic', 'q1', "[network-traffic:dst_port < 1024]")

    # sort by src_port and make sure port 0 comes first
    store.assign('sconns', 'conns', op='sort', by='src_port')
    conns = store.lookup('sconns')
    assert conns[0]['src_port'] == 0
    assert conns[0]['id'] == 'network-traffic--2171d844-d635-4f03-91cc-0a36f1caf3b6_2'


def test_duplicate_identity(fake_bundle_list, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', fake_bundle_list)
    assert store.count('identity') == 1


def test_clobber_viewname(fake_bundle_file_2, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file_2])

    store.extract('conns1', 'network-traffic', 'q1', "[network-traffic:dst_port < 1024]")
    store.extract('conns2', 'network-traffic', 'q1', "[network-traffic:dst_port > 1024]")
    store.rename_view('conns2', 'conns1')  # Clobber conns1

    # conns2 should be no more:
    with pytest.raises(UnknownViewname):
        store.lookup('conns2')


def test_three_ips(one_event_bundle, tmpdir):
    """A single Observation SDO can contain any arbitrary number and type
    of SCOs.  In the case that one type appears multiple times,
    firepit will attempt to mark one as the "primary", or most
    significant, instance by setting an "x_root" attribute to 1 (note
    that this name may change in the future).

    A common case is `ipv4-addr`: if you have a `network-traffic`
    object, then you usally have 2 `ipv4-addr` (or `ipv6-addr`)
    objects.  In that case, firepit will (arbitrarily) pick the object
    referenced as the `src_ref` to be the "primary".

    This test case involves 1 Observation with 3 IP addresses.  One is
    `src_ref`, one is `dst_ref`, and third is...well, ask the QRadar
    people, I guess.

    """
    store = tmp_storage(tmpdir)
    store.cache('q1', [one_event_bundle])

    results = store._query('SELECT value FROM "ipv4-addr" WHERE "x_root" IS NOT NULL')
    rows = results.fetchall()
    assert len(rows) == 1  # There can be only 1!!!
    assert rows[0]['value'] == '10.95.79.130'


def test_finish(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', fake_bundle_file, defer_index=True)
    store.finish()  # No effect with sqlite3
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    urls = store.values('url:value', 'urls')
    assert len(urls) == 14
    assert 'http://www8.example.com/page/176' in urls
    assert 'http://www27.example.com/page/64' not in urls
    store.delete()


def test_grouping_multi_auto(fake_bundle_file, tmpdir):
    # Same test as test_grouping but uses assign_query instead of assign
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])

    store.extract('conns', 'network-traffic', 'q1', "[network-traffic:dst_port < 1024]")
    query = Query()
    query.append(Table('conns'))
    query.append(Group(['src_ref.value']))
    store.assign_query('conns', query)
    srcs = store.values('src_ref.value', 'conns')
    assert srcs

    groups = store.lookup('conns')
    assert groups
    assert 'unique_dst_port' in groups[0].keys()


def test_grouping_multi_agg_1(fake_bundle_file, tmpdir):
    # Same test as test_grouping but uses assign_query instead of assign
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])

    store.extract('conns', 'network-traffic', 'q1', "[network-traffic:dst_port > 0]")
    query = Query([
        Table('conns'),
        Group(['src_ref.value']),
        Aggregation([('SUM', 'number_observed', 'total')]),
        Order([('total', Order.DESC)]),
        Limit(10)
    ])
    store.assign_query('grp_conns', query)

    groups = store.lookup('grp_conns')
    assert groups
    assert 'total' in groups[0].keys()
    assert len(groups) == 10
    expected = [
        ('192.168.90.122', 6),
        ('192.168.160.194', 4),
        ('192.168.57.49', 4),
        ('192.168.70.186', 4),
        ('192.168.104.15', 3),
        ('192.168.132.245', 3),
        ('192.168.152.147', 3),
        ('192.168.156.235', 3),
        ('192.168.203.101', 3),
        ('192.168.0.175', 2),
    ]
    for i, values in enumerate(expected):
        # The order of addrs when count is a tie is not guaranteed
        # assert groups[i]['src_ref.value'] == values[0]
        assert groups[i]['total'] == values[1]


def test_assign_query_1(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])

    store.extract('conns', 'network-traffic', 'q1', "[network-traffic:dst_port > 0]")
    query = Query([
        Table('conns'),
        Order([('src_ref.value', Order.DESC)]),
    ])
    store.assign_query('conns', query)
    srcs = store.values('src_ref.value', 'conns')
    assert srcs[0] > srcs[-1]



import pytest

from firepit.stix20 import stix2sql
from firepit.stix20 import summarize_pattern


def _normalize_ws(s):
    return ' '.join(s.split())


@pytest.mark.parametrize(
    'sco_type, pattern, where', [
        ('ipv4-addr', "[ipv4-addr:value = '9.9.9.9']", "\"value\" = '9.9.9.9'"),
        # Add in some unnecessary yet legal parens
        ('ipv4-addr', "[(ipv4-addr:value = '9.9.9.9')]", "(\"value\" = '9.9.9.9')"),
        # Type doesn't match pattern, so no WHERE clause
        ('process', "[ipv4-addr:value = '9.9.9.9']", ""),
        ('ipv4-addr', "[ipv4-addr:value ISSUBSET '192.168.0.0/16']", "(in_subnet(\"value\", '192.168.0.0/16'))"),
        ('domain-name', "[domain-name:value LIKE 'example.%']", "\"value\" LIKE 'example.%'"),
        ('url',
         "[url:value LIKE 'http://example.%' AND url:value LIKE '%.php']",
         "\"value\" LIKE 'http://example.%' AND \"value\" LIKE '%.php'"),
        ('url',
         "[url:value LIKE 'http://example.%' AND url:value LIKE '%.php' AND url:value LIKE '%foo%']",
         "\"value\" LIKE 'http://example.%' AND \"value\" LIKE '%.php' AND \"value\" LIKE '%foo%'"),
        ('url',
         "[(url:value LIKE 'http://example.%' OR url:value LIKE 'https://example.%') AND url:value LIKE '%foo%']",
         "(\"value\" LIKE 'http://example.%' OR \"value\" LIKE 'https://example.%') AND \"value\" LIKE '%foo%'"),
        # Need to handle reference lists
        ('network-traffic',
         "[network-traffic:protocols[*] = 'tcp']",
         "\"protocols\" LIKE '%tcp%'"),
        ('network-traffic',
         "[network-traffic:protocols[*] != 'tcp']",
         "\"protocols\" NOT LIKE '%tcp%'"),
        ('windows-registry-key',
         "[windows-registry-key:values[*].name = 'foo']",
         "\"values\" LIKE '%\"name\":\"foo\"%'"),
        #TODO: need MATCHES example with PCRE that Python re doesn't support
    ]
)
def test_stix2sql(sco_type, pattern, where):
    assert where == _normalize_ws(stix2sql(pattern, sco_type))


@pytest.mark.parametrize(
    'pattern, expected', [
        ("[ipv4-addr:value = '9.9.9.9']", {"ipv4-addr":{"value"}}),
        ("[url:value LIKE '%blah%']", {"url":{"value"}}),
        ("[process:pid IN (1, 2, 3)]", {"process":{"pid"}}),
        ("[ipv4-addr:value = '9.9.9.9' OR url:value = 'http://example.com/foo']",
         {"ipv4-addr": {"value"}, "url": {"value"}}),
        ("[process:command_line LIKE '% -x' AND process:name = 'foo.exe']",
         {"process": {"command_line", "name"}}),
        ("[url:value LIKE '%blah%'] START t'2017-05-01T18:54:01.000Z' STOP t'2017-05-01T20:27:08.000Z'",
         {"url":{"value"}}),
        ("[network-traffic:dst_port < 10000]", {"network-traffic": {"dst_port"}}),
    ]
)
def test_summarize_pattern(pattern, expected):
    summary = summarize_pattern(pattern)
    for k, v in summary.items():
        print(k, v)
    assert summarize_pattern(pattern) == expected



import os
import pytest

collect_ignore = ['setup.py']


@pytest.fixture
def fake_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, 'test_bundle.json')


@pytest.fixture
def fake_bundle_file_2():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, 'test_bundle_2.json')


@pytest.fixture
def ccoe_bundle():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, 'ccoe_investigator_demo.json')


@pytest.fixture
def fake_csv_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, 'test_procs.csv')


@pytest.fixture
def fake_bundle_list():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return [
        os.path.join(cwd, 'conn_a.json'),
        os.path.join(cwd, 'conn_b.json')
    ]


@pytest.fixture
def one_event_bundle():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, 'one_event.json')



# type: ignore[attr-defined]

"""
Typer-based CLI for testing and experimentation
"""

import csv
import json
import logging
import os
from typing import List

from tabulate import tabulate
import typer

from firepit import get_storage


logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))


app = typer.Typer(
    name="firepit",
    help="Columnar storage for STIX observations",
    add_completion=False,
)


state = {
    'dbname': os.getenv('FIREPITDB', 'stix.db'),
    'session': os.getenv('FIREPITID', 'test-id'),
}


@app.callback()
def main(
    dbname: str = typer.Option(None, help="Path/name of database"),
    session: str = typer.Option(state['session'], help="Session ID to data separation"),
):
    if dbname:
        state['dbname'] = dbname
    if session:
        state['session'] = session


@app.command()
def cache(
    query_id: str = typer.Argument(..., help="An identifier for this set of data"),
    filenames: List[str] = typer.Argument(..., help="STIX bundle files of query results"),
    batchsize: int = typer.Option(2000, help="Max objects to insert per statement"),
):
    """Cache STIX observation data in SQL"""
    db = get_storage(state['dbname'], state['session'])
    if isinstance(filenames, tuple):
        filenames = list(filenames)
    db.cache(query_id, filenames, batchsize)


@app.command()
def extract(
    name: str = typer.Argument(..., help="Name for this new view"),
    sco_type: str = typer.Argument(..., help="SCO type to extract"),
    query_id: str = typer.Argument(..., help="Identifier for cached data to extract from"),
    pattern: str = typer.Argument(..., help="STIX pattern to filter cached data"),
):
    """Create a view of a subset of cached data"""
    db = get_storage(state['dbname'], state['session'])
    db.extract(name, sco_type, query_id, pattern)


@app.command()
def filter(
    name: str = typer.Argument(..., help="Name for this new view"),
    sco_type: str = typer.Argument(..., help="SCO type to extract"),
    source: str = typer.Argument(..., help="Source view"),
    pattern: str = typer.Argument(..., help="STIX pattern to filter cached data"),
):
    """Create a filtered view of a subset of cached data"""
    db = get_storage(state['dbname'], state['session'])
    db.filter(name, sco_type, source, pattern)


@app.command()
def assign(
    name: str = typer.Argument(..., help="Name for this new view"),
    view: str = typer.Argument(..., help="View name to operate on"),
    op: str = typer.Option(..., help="Operation to perform (sort, group, etc.)"),
    by: str = typer.Option(..., help="STIX object path"),
    desc: bool = typer.Option(False, help="Sort descending"),
    limit: int = typer.Option(None, help="Max number of rows to return"),
):
    """Perform an operation on a column and name the result"""
    db = get_storage(state['dbname'], state['session'])
    asc = not desc
    db.assign(name, view, op, by, asc, limit)


@app.command()
def join(
    name: str = typer.Argument(..., help="Name for this new view"),
    left_view: str = typer.Argument(..., help="Left view name to join"),
    left_on: str = typer.Argument(..., help="Column from left view to join on"),
    right_view: str = typer.Argument(..., help="Right view name to join"),
    right_on: str = typer.Argument(..., help="Column from right view to join on"),
):
    """Join two views"""
    db = get_storage(state['dbname'], state['session'])
    db.join(name, left_view, left_on, right_view, right_on)


@app.command()
def lookup(
    name: str = typer.Argument(..., help="View name to look up"),
    limit: int = typer.Option(None, help="Max number of rows to return"),
    offset: int = typer.Option(0, help="Number of rows to skip"),
    format: str = typer.Option('table', help="Output format [table, json]"),
):
    """Retrieve a view"""
    db = get_storage(state['dbname'], state['session'])
    rows = db.lookup(name, limit=limit, offset=offset)
    if format == 'json':
        print(json.dumps(rows, separators=[',', ':']))
    else:
        print(tabulate(rows, headers='keys'))


@app.command()
def values(
    path: str = typer.Argument(..., help="STIX object path to retrieve from view"),
    name: str = typer.Argument(..., help="View name to look up"),
):
    """Retrieve the values of a STIX object path from a view"""
    db = get_storage(state['dbname'], state['session'])
    rows = db.values(path, name)
    for row in rows:
        print(row)


@app.command()
def tables():
    """Get all view/table names"""
    db = get_storage(state['dbname'], state['session'])
    rows = db.tables()
    for row in rows:
        print(row)


@app.command()
def views():
    """Get all view names"""
    db = get_storage(state['dbname'], state['session'])
    rows = db.views()
    for row in rows:
        print(row)


@app.command()
def viewdata(
    views: List[str] = typer.Argument(None, help="Views to merge"),
    format: str = typer.Option('table', help="Output format [table, json]"),
):
    """Get view data for views [default is all views]"""
    db = get_storage(state['dbname'], state['session'])
    rows = db.get_view_data(views)
    if format == 'json':
        print(json.dumps(rows, separators=[',', ':']))
    else:
        print(tabulate(rows, headers='keys'))


@app.command()
def columns(
    name: str = typer.Argument(..., help="View name to look up"),
):
    """Get the columns names of a view/table"""
    db = get_storage(state['dbname'], state['session'])
    rows = db.columns(name)
    for row in rows:
        print(row)


@app.command()
def type(
    name: str = typer.Argument(..., help="View name to look up"),
):
    """Get the SCO type of a view/table"""
    db = get_storage(state['dbname'], state['session'])
    print(db.table_type(name))


@app.command()
def schema(
    name: str = typer.Argument(..., help="View name to look up"),
):
    """Get the schema of a view/table"""
    db = get_storage(state['dbname'], state['session'])
    rows = db.schema(name)
    print(tabulate(rows, headers='keys'))


@app.command()
def count(
    name: str = typer.Argument(..., help="View name to look up"),
):
    """Get the count of rows in a view/table"""
    db = get_storage(state['dbname'], state['session'])
    print(db.count(name))


@app.command()
def delete(
):
    """Delete STIX observation data in SQL"""
    db = get_storage(state['dbname'], state['session'])
    db.delete()


@app.command()
def sql(
    statement: str = typer.Argument(..., help="View name to look up"),
):
    """Run a SQL statement on the database [DANGEROUS!]"""
    db = get_storage(state['dbname'], state['session'])
    rows = db._execute(statement)
    if rows:
        print(tabulate(rows, headers='keys'))


@app.command()
def set_appdata(
    name: str = typer.Argument(..., help="View name"),
    data: str = typer.Argument(..., help="Data (string)"),
):
    """Set the app-specific data for a view"""
    db = get_storage(state['dbname'], state['session'])
    db.set_appdata(name, data)


@app.command()
def get_appdata(
    name: str = typer.Argument(..., help="View name"),
):
    """Get the app-specific data for a view"""
    db = get_storage(state['dbname'], state['session'])
    print(db.get_appdata(name))


@app.command()
def load(
    name: str = typer.Argument(..., help="Name for this new view"),
    sco_type: str = typer.Option(None, help="SCO type of data to load"),
    query_id: str = typer.Option(None, help="An identifier for this set of data"),
    preserve_ids: str = typer.Option(True, help="Use IDs in the data"),
    filename: str = typer.Argument(..., help="Data file to load (JSON only)"),
):
    """Cache STIX observation data in SQL"""
    db = get_storage(state['dbname'], state['session'])
    try:
        with open(filename, 'r') as fp:
            data = json.load(fp)
    except json.decoder.JSONDecodeError:
        with open(filename, 'r') as fp:
            data = list(csv.DictReader(fp))
    db.load(name, data, sco_type, query_id, preserve_ids)


@app.command()
def reassign(
    name: str = typer.Argument(..., help="Name for this new view"),
    filename: str = typer.Argument(..., help="Data file to load (JSON only)"),
):
    """Update/replace STIX observation data in SQL"""
    db = get_storage(state['dbname'], state['session'])
    with open(filename, 'r') as fp:
        data = json.load(fp)
    db.reassign(name, data)


@app.command()
def merge(
    name: str = typer.Argument(..., help="Name for this new view"),
    views: List[str] = typer.Argument(..., help="Views to merge"),
):
    """Merge 2 or more views into a new view"""
    db = get_storage(state['dbname'], state['session'])
    db.merge(name, views)


@app.command()
def remove(
    name: str = typer.Argument(..., help="Name of view to remove"),
):
    """Remove a view"""
    db = get_storage(state['dbname'], state['session'])
    db.remove_view(name)


@app.command()
def rename(
    oldname: str = typer.Argument(..., help="Name of view to rename"),
    newname: str = typer.Argument(..., help="New name of view to rename"),
):
    """Rename a view"""
    db = get_storage(state['dbname'], state['session'])
    db.rename_view(oldname, newname)


if __name__ == "__main__":
    app()



import os
import pytest

from firepit.exceptions import IncompatibleType
from firepit.exceptions import InvalidAttr
from firepit.exceptions import InvalidStixPath
from firepit.exceptions import InvalidViewname
from firepit.exceptions import StixPatternError
from firepit.exceptions import UnexpectedError
from firepit.query import Filter
from firepit.query import Group
from firepit.query import Predicate
from firepit.query import Query
from firepit.query import Table

from .helpers import tmp_storage


@pytest.fixture
def invalid_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, 'test_error_bundle.json')


def test_local(invalid_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [invalid_bundle_file])


def test_extract_bad_stix_pattern(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    with pytest.raises(StixPatternError):
        store.extract('junk', 'ipv4-addr', 'q1', "whatever")


def test_filter_bad_stix_pattern(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    with pytest.raises(StixPatternError):
        store.filter('junk', 'url', 'urls', "value = 'http://www26.example.com/page/176'")


def test_filter_bad_input_view(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    with pytest.raises(InvalidViewname):
        store.filter('junk', 'url', 'urls OR 1', "[url:value = 'http://www26.example.com/page/176']")


def test_sqli_1(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    with pytest.raises(InvalidViewname):
        store.lookup('urls" UNION ALL SELECT * FROM "q1_url')


def test_sqli_2(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    with pytest.raises(InvalidAttr):
        store.values('url:fake.path', 'urls')
    with pytest.raises(InvalidStixPath):
        store.values('value" FROM "q1_ipv4-addr" UNION ALL SELECT "value', 'urls')


def test_sqli_3(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")

    res = store.load('test_urls', [
        {
            'type': 'url',
            'value': 'http://www26.example.com/page/176',
            'risk': 'high',
        },
        {
            'type': 'url',
            'value': 'http://www67.example.com/page/264',
            'risk': 'high',
        }
    ])

    with pytest.raises(InvalidViewname):
        store.join('sqli" AS SELECT * FROM "q1_url"; CREATE VIEW "marked',
                   'urls', 'value', 'test_urls', 'value')


def test_query_sqli_table(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    with pytest.raises(InvalidViewname):
        query = Query([Table('urls; select * from url; --')])
        store.run_query(query)


def test_query_sqli_predicate(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    query = Query('url')
    count = len(store.run_query(query).fetchall())
    assert count  # Make sure test is valid
    query = Query([
        Table('urls'),
        # This will not raise, but underlying SQL driver should prevent injection
        Filter([Predicate('value', '=', '1; select * from url; --')])
    ])
    result = store.run_query(query).fetchall()
    assert len(result) == 0  # If injection succeeded, we'd get len(result) == count


def test_empty_results(fake_bundle_file, tmpdir):
    """Look for finding objects that aren't there"""
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('my_findings', 'x-ibm-finding', 'q1', "[x-ibm-finding:name = 'Whatever']")
    findings = store.lookup('my_findings')
    assert findings == []


def test_lookup_bad_columns(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    with pytest.raises(InvalidAttr):
        store.lookup('urls', cols="1; select * from urls; --")


def test_lookup_bad_offset(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    with pytest.raises(TypeError):
        store.lookup('urls', offset="1; select * from urls; --")


def test_bad_groupby(fake_bundle_file, fake_csv_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('users', 'user-account', 'q1', "[ipv4-addr:value LIKE '10.%']")
    with pytest.raises(InvalidStixPath):
        store.assign('grouped_users', 'users', op='group',
                     by='1,extractvalue(0x0a,concat(0x0a,(select database())))--')


def test_assign_bad_columns(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    with pytest.raises(InvalidStixPath):
        store.assign('sorted', 'urls', op='sort',
                     by='value LIMIT 1; SELECT * FROM "urls"')


def test_sort_bad_limit(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    with pytest.raises(TypeError):
        store.assign('sorted', 'urls', op='sort', by='value', limit='1; SELECT 1; --')


def test_merge_fail(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)

    store.cache('test-bundle', [fake_bundle_file])
    store.extract('urls', 'url', 'test-bundle', "[url:value LIKE '%page/1%']")
    store.extract('ips', 'ipv4-addr', 'test-bundle', "[ipv4-addr:value != '8.8.8.8']")

    with pytest.raises(IncompatibleType):
        store.merge('merged', ['urls', 'ips'])


def test_assign_query_sqli(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    query = Query([
        Table('url'),
        Filter([Predicate('value', '=', '1; select * from url; --')])
    ])
    with pytest.raises(UnexpectedError):
        store.assign_query('urls', query)


def test_assign_query_sqli_group(fake_bundle_file, tmpdir):
    store = tmp_storage(tmpdir)
    store.cache('q1', [fake_bundle_file])
    store.extract('conns', 'network-traffic', 'q1', "[network-traffic:dst_port < 1024]")
    with pytest.raises(InvalidStixPath):
        query = Query([
            Table('conns'),
            Group(['src_ref.value where 1=2; select * from identity; --'])
        ])



import uuid
import pytest

from firepit.exceptions import InvalidStixPath
from firepit.exceptions import InvalidViewname
from firepit.validate import validate_name
from firepit.validate import validate_path


@pytest.mark.parametrize(
    'name, expected', [
        ('foo', True),
        ('[*]', False),
        ('__tmp_6668fcc6300f40e39c255c6573d79180', True),
        ('__tmp_' + uuid.uuid4().hex, True),
        ('foo;', False),
        ('foo; --', False),
    ]
)
def test_validate_name(name, expected):
    if expected:
        validate_name(name)
    else:
        with pytest.raises(InvalidViewname):
            validate_name(name)


@pytest.mark.parametrize(
    'path, expected', [
        ('foo', True),
        ('things[*]', True),
        ('one.two', True),
        ('foo;', False),
        ('foo; --', False),
        ('foo."bar"', False),
        ('ipv4_addr:value', False),
        ("hashes.'SHA-256'", True),
        ("values[*].name", True),
        ("extensions.'http-request-ext'.request_headers.'Content-Type'", True),
        ("ipv4-addr:value", True),
        ("file:hashes.'SHA-1'", True),
        ("file:hashes.IMPHASH", True),
        ("windows-registry-key:values[*].data", True),
        ("network-traffic:protocols[*]", True),
    ]
)
def test_validate_path(path, expected):
    if expected:
        validate_path(path)
    else:
        with pytest.raises(InvalidStixPath):
            validate_path(path)



import os
import pytest

from firepit.dataframe import DataFrame

from .helpers import tmp_storage


@pytest.fixture
def store(tmpdir):
    cwd = os.path.dirname(os.path.abspath(__file__))
    path = os.path.join(cwd, 'test_bundle.json')
    store = tmp_storage(tmpdir)
    store.cache('q1', path)
    store.extract('urls', 'url', 'q1', "[url:value LIKE '%page/1%']")
    store.extract('conns', 'network-traffic', 'q1', "[ipv4-addr:value LIKE '10.%']")
    return store


@pytest.mark.parametrize(
    'view, column, op, expected', [
        ('urls', 'value', 'count', 14),
        ('urls', 'value', 'nunique', 14),
        ('urls', 'value', 'min', 'http://www11.example.com/page/108'),
        ('urls', 'value', 'max', 'http://www91.example.com/page/104'),
        ('conns', 'src_port', 'count', 100),
        ('conns', 'src_port', 'min', 49434),
        ('conns', 'src_port', 'max', 65528),
        ('conns', 'dst_port', 'sum', 130350),
        ('conns', 'src_port', 'nunique', 100),
        ('conns', 'dst_port', 'nunique', 5),
    ]
)
def test_series(store, view, column, op, expected):
    df = DataFrame(view, store)
    if op == 'count':
        assert df[column].count() == expected
    elif op == 'nunique':
        assert df[column].nunique() == expected
    elif op == 'min':
        assert df[column].min() == expected
    elif op == 'max':
        assert df[column].max() == expected
    elif op == 'sum':
        assert df[column].sum() == expected


@pytest.mark.parametrize(
    'view, columns', [
        ('urls', ['type', 'value', 'x_root', 'id', 'x_contained_by_ref',
                  'first_observed', 'last_observed', 'number_observed']),
        ('conns', ['type', 'start', 'end', 'src_ref.type', 'src_ref.value',
                   'src_ref.x_root', 'src_ref.id', 'dst_ref.type', 'dst_ref.value',
                   'dst_ref.id', 'src_port', 'dst_port','protocols', 'x_root',
                   'id', 'x_contained_by_ref', 'first_observed', 'last_observed',
                   'number_observed'])
    ]
)
def test_dataframe_columns(store, view, columns):
    df = DataFrame(view, store)
    assert set(df.columns) == set(columns)


@pytest.mark.parametrize(
    'view, count', [
        ('urls', 14),
        ('conns', 100)
    ]
)
def test_dataframe_columns(store, view, count):
    df = DataFrame(view, store)
    index = df.index
    assert index[0] == 0
    assert len(index) == count


@pytest.mark.parametrize(
    'view, shape', [
        ('urls', (8, 14)),
        ('conns', (19, 100))
    ]
)
def test_dataframe_shape(store, view, shape):
    df = DataFrame(view, store)
    assert df.shape == shape    



class InvalidAttr(Exception):
    def __init__(self, msg):
        self.message = msg

    def __str__(self):
        return f"{self.message}"


class InvalidObject(Exception):
    def __init__(self, msg):
        self.message = msg

    def __str__(self):
        return f"{self.message}"


class StixPatternError(Exception):
    def __init__(self, stix):
        self.stix = stix

    def __str__(self):
        return f"{self.stix}"


class InvalidViewname(Exception):
    pass


class InvalidStixPath(Exception):
    pass


class IncompatibleType(Exception):
    pass


class UnknownViewname(Exception):
    pass


class DuplicateTable(Exception):
    pass


class UnexpectedError(Exception):
    pass



import logging
import re
import uuid

import orjson
import ujson

from firepit import raft
from firepit.exceptions import IncompatibleType
from firepit.exceptions import InvalidAttr
from firepit.exceptions import InvalidObject
from firepit.exceptions import StixPatternError
from firepit.props import auto_agg
from firepit.props import auto_agg_tuple
from firepit.props import primary_prop
from firepit.query import Aggregation
from firepit.query import Group
from firepit.query import Table
from firepit.splitter import SplitWriter
from firepit.stix20 import stix2sql
from firepit.validate import validate_name
from firepit.validate import validate_path

logger = logging.getLogger(__name__)


def _transform(filename):
    for obj in raft.get_objects(filename, ['identity', 'observed-data']):
        if obj['type'] == 'observed-data':
            obj['x_stix'] = ujson.dumps(obj)  # raft.preserve
            obj = raft.invert(obj)
            obj = raft.markroot(obj)  # Also does raft.makeid now
            observables = obj.get('objects', {})
            # Inlined below: obj = raft.nest(obj)
            object_refs = []
            for obs_orig in observables.values():
                object_refs.append(raft._resolve(obs_orig, observables))
            obj['objects'] = object_refs
            objs = raft.promote(obj)
            # Inlined below: raft.normalize
            objs = [raft.json_normalize(obj, flat_lists=False) for obj in objs]
            for obj in objs:
                yield obj
        else:
            yield obj


def infer_type(key, value):
    if key == 'id':
        rtype = 'TEXT UNIQUE'
    elif key == 'number_observed':
        rtype = 'NUMERIC'
    elif isinstance(value, int):
        rtype = 'NUMERIC'
    elif isinstance(value, float):
        rtype = 'REAL'
    elif isinstance(value, list):
        rtype = 'TEXT'
    else:
        rtype = 'TEXT'
    return rtype


class SqlStorage:
    def __init__(self):
        self.connection = None  # Python DB API connection object
        self.placeholder = '%s'  # Derived class can override this

        # Functions to use for min/max text.  It can vary - sqlite3
        # uses MIN/MAX, postgresql uses LEAST/GREATEST
        self.text_min = 'MIN'
        self.text_max = 'MAX'

        # Function that returns first non-null arg_type
        self.ifnull = 'IFNULL'

        # Python-to-SQL type mapper
        self.infer_type = infer_type

    def _get_writer(self, **kwargs):
        """Get a DB inserter object"""
        # This is DB-specific
        raise NotImplementedError('SqlStorage._get_writer')

    def _initdb(self, cursor):
        """Do some initial DB setup"""
        stmt = ('CREATE TABLE IF NOT EXISTS "__symtable" '
                '(name TEXT, type TEXT, appdata TEXT);')
        self._execute(stmt, cursor)
        stmt = ('CREATE TABLE IF NOT EXISTS "__queries" '
                '(sco_id TEXT, query_id TEXT);')
        self._execute(stmt, cursor)
        self.connection.commit()
        cursor.close()

    def _new_name(self, cursor, name, sco_type):
        stmt = ('INSERT INTO "__symtable" (name, type)'
                f' VALUES ({self.placeholder}, {self.placeholder});')
        cursor.execute(stmt, (name, sco_type))

    def _drop_name(self, cursor, name):
        stmt = f'DELETE FROM "__symtable" WHERE name = {self.placeholder};'
        cursor.execute(stmt, (name,))

    def _execute(self, statement, cursor=None):
        """Private wrapper for logging SQL statements"""
        logger.debug('Executing statement: %s', statement)
        if not cursor:
            cursor = self.connection.cursor()
        cursor.execute(statement)
        return cursor

    def _command(self, cmd, cursor=None):
        """Private wrapper for logging SQL commands"""
        logger.debug('Executing command: %s', cmd)
        if not cursor:
            cursor = self.connection.cursor()
        cursor.execute(cmd)
        self.connection.commit()

    def _query(self, query, values=None, cursor=None):
        """Private wrapper for logging SQL query"""
        logger.debug('Executing query: %s', query)
        if not cursor:
            cursor = self.connection.cursor()
        if not values:
            values = ()
        cursor.execute(query, values)
        self.connection.commit()
        return cursor

    def _select(self, tvname, cols="*", sortby=None, groupby=None,
                ascending=True, limit=None, offset=None, where=None):
        """Generate a SELECT query on table or view `tvname`"""
        # TODO: Deprecate this in favor of query module
        validate_name(tvname)
        if cols != "*":
            cols = ", ".join([f'"{col}"' if not col.startswith("'") else col for col in cols])

        stmt = f'SELECT {cols} FROM "{tvname}"'
        if where:
            stmt += f' WHERE {where}'
        if groupby:
            validate_path(groupby)

            # For grouping, we need to aggregate data in the columns.
            aggs = [
                'MIN("type") as "type"',
                f'"{groupby}"',
            ]
            sco_type = self.table_type(tvname)
            for col in self.schema(tvname):
                # Don't aggregate the column we used for grouping
                if col['name'] == groupby:
                    continue
                agg = auto_agg(sco_type, col['name'], col['type'])
                if agg:
                    aggs.append(agg)
            group_cols = ', '.join(aggs)
            stmt = f'SELECT {group_cols} from "{tvname}"'
            stmt += f' GROUP BY "{groupby}"'
        if sortby:
            validate_path(sortby)
            stmt += f' ORDER BY "{sortby}" ' + ('ASC' if ascending else 'DESC')
        if limit:
            if not isinstance(limit, int):
                raise TypeError('LIMIT must be an integer')
            stmt += f' LIMIT {limit}'
        if offset:
            if not isinstance(offset, int):
                raise TypeError('LIMIT must be an integer')
            stmt += f' OFFSET {offset}'
        return stmt

    def _create_table(self, tablename, columns):
        stmt = f'CREATE TABLE "{tablename}" ('
        stmt += ','.join([f'"{colname}" {coltype}' for colname, coltype in columns.items()])
        stmt += ');'
        logger.debug('_create_table: "%s"', stmt)
        cursor = self._execute(stmt)
        if 'x_contained_by_ref' in columns:
            self._execute(f'CREATE INDEX "{tablename}_obs" ON "{tablename}" ("x_contained_by_ref");', cursor)
        self.connection.commit()
        cursor.close()

    def _add_column(self, tablename, prop_name, prop_type):
        stmt = f'ALTER TABLE "{tablename}" ADD COLUMN "{prop_name}" {prop_type};'
        logger.debug('new_property: "%s"', stmt)
        self._execute(stmt)

    def _create_view(self, viewname, select, sco_type, deps=None, cursor=None):
        # This is DB-specific
        raise NotImplementedError('Storage._create_view')

    def _get_view_def(self, viewname):
        # This is DB-specific
        raise NotImplementedError('Storage._get_view_def')

    def _is_sql_view(self, name, cursor=None):
        ## This is DB-specific
        raise NotImplementedError('Storage._is_sql_view')

    def _extract(self, viewname, sco_type, tablename, pattern, query_id=None):
        """Extract rows from `tablename` to create view `viewname`"""
        validate_name(viewname)
        validate_name(tablename)
        try:
            where = stix2sql(pattern, sco_type) if pattern else None
        except Exception as e:
            logger.error('%s', e)
            raise StixPatternError(pattern) from e
        if query_id:
            clause = f"query_id = '{query_id}'"
            if where:
                where = f"{clause} AND ({where})"
            else:
                where = clause

        # Need to convert viewname from identifier to string, so use single quotes
        namestr = f"'{viewname}'"
        cursor = self._execute('BEGIN;')
        select = (f'SELECT * FROM "{sco_type}" WHERE "id" IN'
                  f' (SELECT "{sco_type}".id FROM "{sco_type}"'
                  f'  INNER JOIN __queries ON "{sco_type}".id = __queries.sco_id'
                  f'  WHERE {where});')

        cursor = self._create_view(viewname, select, sco_type, deps=[tablename], cursor=cursor)
        self.connection.commit()
        cursor.close()

    def _get_excluded(self, colnames, tablename):
        excluded = []
        for col in colnames:
            if col == 'first_observed':
                excluded.append(f'first_observed = {self.text_min}("{tablename}".first_observed, EXCLUDED.first_observed)')
            elif col == 'last_observed':
                excluded.append(f'last_observed = {self.text_max}("{tablename}".last_observed, EXCLUDED.last_observed)')
            elif col == 'number_observed':
                excluded.append(f'number_observed = "{tablename}".number_observed + EXCLUDED.number_observed')
            elif col == 'id':
                continue
            else:
                excluded.append(f'"{col}" = EXCLUDED."{col}"')
        return ', '.join(excluded)

    def upsert(self, cursor, tablename, obj, query_id, schema):
        colnames = list(schema.keys())
        excluded = self._get_excluded(colnames, tablename)
        valnames = ', '.join([f'"{x}"' for x in colnames])
        placeholders = ', '.join([self.placeholder] * len(obj))
        stmt = f'INSERT INTO "{tablename}" ({valnames}) VALUES ({placeholders})'
        if 'id' in colnames:
            stmt += f' ON CONFLICT (id) DO UPDATE SET {excluded}'
        values = tuple([str(orjson.dumps(value), 'utf-8')
                        if isinstance(value, list) else value for value in obj])
        logger.debug('_upsert: "%s"', stmt)
        cursor.execute(stmt, values)

        if query_id:
            # Now add to query table as well
            idx = colnames.index('id')
            stmt = (f'INSERT INTO "__queries" (sco_id, query_id)'
                    f' VALUES ({self.placeholder}, {self.placeholder})')
            cursor.execute(stmt, (obj[idx], query_id))

    def upsert_many(self, cursor, tablename, objs, query_id, schema):
        for obj in objs:
            self.upsert(cursor, tablename, obj, query_id, schema)

    def cache(self, query_id, bundles, batchsize=2000, **kwargs):
        """Cache the result of a query/dataset

        Takes the `observed-data` SDOs from `bundles` and "flattens"
        them, splits out SCOs by type, and inserts into a database
        with 1 table per type.

        Accepts some keyword args for runtime options, some of which
        may depend on what database type is in use (e.g. sqlite3,
        postgresql, ...)

        Args:

          query_id (str): a unique identifier for this set of bundles

          bundles (list): STIX bundles (either in-memory Python objects or filename paths)

          batchsize (int): number of objects to insert in 1 batch (defaults to 2000)

        """
        logger.debug('Caching %s', query_id)

        if not isinstance(bundles, list):
            bundles = [bundles]

        writer = self._get_writer(**kwargs)
        splitter = SplitWriter(writer, batchsize=batchsize, query_id=str(query_id))

        # walk the bundles and figure out all the columns
        for bundle in bundles:
            if isinstance(bundle, str):
                logger.debug('- Caching %s', bundle)
            for obj in _transform(bundle):
                splitter.write(obj)
        splitter.close()

    def assign(self, viewname, on, op=None, by=None, ascending=True, limit=None):
        """
        Perform (unary) operation `op` on `on` and store result as `viewname`
        """
        validate_name(viewname)
        if by:
            validate_path(by)
            _, _, by = by.rpartition(':')
        if op == 'sort':
            stmt = self._select(
                on, sortby=by, ascending=ascending, limit=limit)
        elif op == 'group':
            stmt = self._select(on, groupby=by)
        sco_type = self.table_type(on)
        cursor = self._create_view(viewname, stmt, sco_type, deps=[on])
        self.connection.commit()
        cursor.close()

    def load(self, viewname, objects, sco_type=None, query_id=None, preserve_ids=True):
        """Import `objects` as type `sco_type` and store as `viewname`"""
        validate_name(viewname)
        if not query_id:
            # Look inside data
            if 'query_id' in objects[0]:
                query_id = objects[0]['query_id']
            else:
                query_id = str(uuid.uuid4())
        writer = self._get_writer(query_id=query_id)
        splitter = SplitWriter(writer, batchsize=1000, query_id=str(query_id))

        for obj in objects:
            if not sco_type:
                # objects MUST be dicts with a type
                if 'type' not in obj:
                    raise InvalidObject('missing `type`')
                sco_type = obj['type']
            if isinstance(obj, str):
                obj = {'type': sco_type, primary_prop(sco_type): obj}
            elif not isinstance(obj, dict):
                raise InvalidObject('Unknown data format')
            if 'id' not in obj or not preserve_ids:
                obj['id'] = sco_type + '--' + str(uuid.uuid4())
            if 'type' not in obj:
                obj['type'] = sco_type
            splitter.write(obj)
        splitter.close()

        self.extract(viewname, sco_type, query_id, '')

        return sco_type

    def reassign(self, viewname, objects):
        """Replace `objects` (or insert them if they're not there)"""
        validate_name(viewname)
        # TODO: ensure viewname exists?  Do we care?

        # Ignore it if objects is empty
        if not objects:
            return

        cursor = self._execute('BEGIN;')
        if 'id' not in objects[0]:
            # Maybe it's aggregates?  Do "copy-on-write"
            self._execute(f'DROP VIEW IF EXISTS "{viewname}"', cursor)
            columns = list(objects[0].keys())
            schema = {}
            for col in columns:
                schema[col] = self.infer_type(col, objects[0][col])
            self._create_table(viewname, schema)
            records = [[obj.get(col) for col in columns] for obj in objects]
            self.upsert_many(cursor, viewname, records, None, schema)
            viewdef = self._select(viewname)
        else:
            writer = self._get_writer()
            splitter = SplitWriter(writer, batchsize=1000, replace=True)
            sco_type = None
            for obj in objects:
                if 'type' not in obj:
                    raise InvalidObject('missing `type`')
                elif not isinstance(obj, dict):
                    raise InvalidObject('Unknown data format')
                if not sco_type:
                    sco_type = obj['type']
                if 'id' not in obj:
                    raise InvalidObject('missing `id`')
                splitter.write(obj)
            splitter.close()
            viewdef = self._get_view_def(viewname)
            self._execute(f'DROP VIEW IF EXISTS "{viewname}"', cursor)

            # Recreate view
            self._execute(f'CREATE VIEW "{viewname}" AS {viewdef}', cursor)

        self.connection.commit()

    def update(self, objects, query_id=None):
        """Update `objects`"""
        writer = self._get_writer()
        splitter = SplitWriter(writer, batchsize=1000, replace=True)
        for obj in objects:
            if 'type' not in obj:
                raise InvalidObject('missing `type`')
            elif not isinstance(obj, dict):
                raise InvalidObject('Unknown data format')
            if 'id' not in obj:
                raise InvalidObject('missing `id`')
            splitter.write(obj)
        splitter.close()

    def join(self, viewname, l_var, l_on, r_var, r_on):
        """Join vars `l_var` and `r_var` and store result as `viewname`"""
        validate_name(viewname)
        validate_name(l_var)
        validate_name(r_var)
        validate_path(l_on)
        validate_path(r_on)
        l_cols = self.columns(l_var)
        r_cols = self.columns(r_var)
        l_type, _, l_on = l_on.rpartition(':')
        r_type, _, r_on = r_on.rpartition(':')
        cols = set()
        for col in l_cols:
            cols.add(f'{self.ifnull}({l_var}."{col}", {r_var}."{col}") AS "{col}"')
        for col in r_cols:
            # Only add if not already added from left
            if col not in l_cols:
                cols.add(f'{r_var}."{col}" as "{col}"')
        scols = ', '.join(cols)
        stmt = (f'SELECT {scols} FROM'
                f' {l_var} INNER JOIN {r_var}'
                f' ON {l_var}."{l_on}" = {r_var}."{r_on}"')
        sco_type = self.table_type(l_var)
        cursor = self._create_view(viewname, stmt, sco_type, deps=[l_var, r_var])
        self.connection.commit()
        cursor.close()

    def extract(self, viewname, sco_type, query_id, pattern):
        """
        Extract all `sco_type` object from the results of `query_id` and
        store as `viewname`

        """
        validate_name(viewname)
        logger.debug('Extract %s as %s from %s with %s',
                     sco_type, viewname, query_id, pattern)
        self._extract(viewname, sco_type, sco_type, pattern, query_id)

    def filter(self, viewname, sco_type, input_view, pattern):
        """
        Extract all `sco_type` object from `input_view` and store as
        `viewname`

        """
        validate_name(viewname)
        validate_name(input_view)
        logger.debug('Filter %s as %s from %s with %s',
                     sco_type, viewname, input_view, pattern)
        slct = self._get_view_def(input_view)
        try:
            where = stix2sql(pattern, sco_type) if pattern else None
        except Exception as e:
            logger.error('%s', e)
            raise StixPatternError(pattern) from e
        slct = f'SELECT * FROM ({slct}) AS tmp'
        if where:
            slct += f' WHERE {where}'
        cursor = self._create_view(viewname, slct, sco_type, deps=[input_view])
        self.connection.commit()
        cursor.close()

    def lookup(self, viewname, cols="*", limit=None, offset=None):
        """Get the value of `viewname`"""
        validate_name(viewname)
        if cols != "*":
            dbcols = self.columns(viewname)
            cols = cols.replace(" ", "").split(",")
            for col in cols:
                if col not in dbcols:
                    raise InvalidAttr(f"{col}")
        stmt = self._select(viewname, cols=cols, limit=limit, offset=offset)
        cursor = self._query(stmt)
        return cursor.fetchall()

    def values(self, path, viewname):
        """Get the values of STIX object path `path` (a column) from `viewname`"""
        validate_path(path)
        validate_name(viewname)
        _, _, column = path.rpartition(':')
        if column not in self.columns(viewname):
            raise InvalidAttr(path)
        stmt = f'SELECT "{column}" FROM "{viewname}"'
        cursor = self._query(stmt)
        result = cursor.fetchall()
        return [row[column] for row in result]

    def count(self, viewname):
        """Get the count of objects (rows) in `viewname`"""
        validate_name(viewname)
        stmt = f'SELECT COUNT(*) FROM "{viewname}"'
        cursor = self._query(stmt)
        res = cursor.fetchone()
        return list(res.values())[0] if res else 0

    def tables(self):
        """Get all table names"""
        # This is DB-specific
        raise NotImplementedError('Storage.tables')

    def types(self):
        """Get all table names that correspond to SCO types"""
        # This is DB-specific
        raise NotImplementedError('Storage.types')

    def views(self):
        """Get all view names"""
        stmt = 'SELECT name FROM __symtable'
        cursor = self._query(stmt)
        result = cursor.fetchall()
        return [row['name'] for row in result]

    def table_type(self, viewname):
        """Get the SCO type for table/view `viewname`"""
        validate_name(viewname)
        #stmt = f'SELECT "type" FROM "{viewname}" WHERE "type" IS NOT NULL LIMIT 1;'
        stmt = f'SELECT "type" FROM "__symtable" WHERE name = {self.placeholder};'
        cursor = self._query(stmt, (viewname,))
        res = cursor.fetchone()
        return list(res.values())[0] if res else None

    def columns(self, viewname):
        """Get the column names (properties) of `viewname`"""
        # This is DB-specific
        raise NotImplementedError('Storage.columns')

    def schema(self, viewname):
        """Get the schema (names and types) of `viewname`"""
        # This is DB-specific
        raise NotImplementedError('Storage.schema')

    def delete(self):
        """Delete ALL data in this store"""
        # This is DB-specific
        raise NotImplementedError('Storage.delete')

    def set_appdata(self, viewname, data):
        """Attach app-specific data to a viewname"""
        validate_name(viewname)
        stmt = (f'UPDATE "__symtable" SET appdata = {self.placeholder}'
                f' WHERE name = {self.placeholder};')
        values = (data, viewname)
        cursor = self._query(stmt, values=values)
        cursor.close()

    def get_appdata(self, viewname):
        """Retrieve app-specific data for a viewname"""
        validate_name(viewname)
        stmt = f'SELECT appdata FROM "__symtable" WHERE name = {self.placeholder};'
        values = (viewname,)
        cursor = self._query(stmt, values)
        res = cursor.fetchone()
        cursor.close()
        if not res:
            return None
        if 'appdata' in res:
            return res['appdata']
        return res[0]

    def get_view_data(self, viewnames=None):
        """Retrieve information about one or more viewnames"""
        if viewnames:
            placeholders = ', '.join([self.placeholder] * len(viewnames))
            stmt = f'SELECT * FROM "__symtable" WHERE name IN ({placeholders});'
            values = tuple(viewnames)
        else:
            stmt = 'SELECT * FROM "__symtable";'
            values = None
        cursor = self._query(stmt, values)
        res = cursor.fetchall()
        cursor.close()
        return res

    def run_query(self, query):
        query_text, query_values = query.render(self.placeholder)
        return self._query(query_text, query_values)

    def merge(self, viewname, input_views):
        validate_name(viewname)
        selects = []
        types = set()
        for name in input_views:
            validate_name(name)
            types.add(self.table_type(name))
            selects.append(self._get_view_def(name))
        if len(types) > 1:
            raise IncompatibleType('cannot merge types ' + ', '.join(types))
        stmt = ' UNION '.join(selects)
        sco_type = self.table_type(input_views[0])
        cursor = self._create_view(viewname, stmt, sco_type, deps=input_views)
        self.connection.commit()
        cursor.close()

    def remove_view(self, viewname):
        """Remove view `viewname`"""
        validate_name(viewname)
        cursor = self._execute('BEGIN;')
        self._execute(f'DROP VIEW IF EXISTS "{viewname}";', cursor)
        self._drop_name(cursor, viewname)
        self.connection.commit()
        cursor.close()

    def rename_view(self, oldname, newname):
        """Rename view `oldname` to `newname`"""
        validate_name(oldname)
        validate_name(newname)
        view_type = self.table_type(oldname)
        view_def = self._get_view_def(oldname)
        cursor = self._execute('BEGIN;')

        # Need to remove `newname` if it already exists
        self._execute(f'DROP VIEW IF EXISTS "{newname}";', cursor)
        self._drop_name(cursor, newname)

        # Now do the rename
        qry = re.sub(f'var = \'{oldname}\'',  # This is an ugly hack
                     f'var = \'{newname}\'',
                     view_def)
        self._create_view(newname, qry, view_type, cursor=cursor)
        self._execute(f'DROP VIEW IF EXISTS "{oldname}"', cursor)
        self._drop_name(cursor, oldname)
        self._new_name(cursor, newname, view_type)

        self.connection.commit()
        cursor.close()

    def finish(self):
        """Do any DB-specific post-caching/insertion activity, such as indexing"""
        # This is a DB-specific hook, but by default we'll do nothing
        pass

    def assign_query(self, viewname, query):
        """
        Create a new view `viewname` defined by `query`
        """
        # Deduce SCO type and "deps" of viewname from query
        deps = []
        sco_type = ''  # TODO: what to use for unknown type?
        group_cols = []
        found_agg = False
        for stage in query.stages:
            if isinstance(stage, Table):
                # Assume first Table?  Might not work for complex queries or joins
                on = stage.name
                sco_type = self.table_type(on)
                deps = [on]
            elif isinstance(stage, Aggregation):
                found_agg = True
            elif isinstance(stage, Group):
                group_cols = stage.cols

        # if no aggs supplied, do "auto aggregation"
        if group_cols and not found_agg and sco_type:
            aggs = []
            for col in self.schema(sco_type):  #viewname):
                # Don't aggregate the columns we used for grouping
                if col['name'] in group_cols:
                    continue
                agg = auto_agg_tuple(sco_type, col['name'], col['type'])
                if agg:
                    aggs.append(agg)
            query.append(Aggregation(aggs))

        query_text, query_values = query.render('{}')
        stmt = query_text.format(query_values)
        logger.debug('assign_query: %s', stmt)
        cursor = self._create_view(viewname, stmt, sco_type, deps=deps)
        self.connection.commit()
        cursor.close()



import pytest

from firepit.query import Aggregation
from firepit.query import Count
from firepit.query import CountUnique
from firepit.query import Filter
from firepit.query import Group
from firepit.query import InvalidComparisonOperator
from firepit.query import InvalidQuery
from firepit.query import Join
from firepit.query import Limit
from firepit.query import Offset
from firepit.query import Order
from firepit.query import Predicate
from firepit.query import Projection
from firepit.query import Query
from firepit.query import Table
from firepit.query import Unique


@pytest.mark.parametrize(
    'lhs, op, rhs, expected_len', [
        ('foo', '=', 99, 1),
        ('bar', '>=', 99, 1),
        ('baz', 'LIKE', '%blah%', 1),
    ]
)
def test_predicate(lhs, op, rhs, expected_len):
    p1 = Predicate(lhs, op, rhs)
    text = p1.render('?')
    assert str(rhs) not in text
    assert '?' in text
    assert len(p1.values) == expected_len
    assert p1.values[0] == rhs


@pytest.mark.parametrize(
    'lhs, op, rhs, expected_text', [
        ('foo', '=', 'null', '("foo" IS NULL)'),
        ('bar', '!=', 'NULL', '("bar" IS NOT NULL)'),
        ('baz[*]', '=', 'NULL', '("baz" IS NULL)'),
        ('next.name[*]', '!=', 'null', '("next.name" IS NOT NULL)'),
    ]
)
def test_predicate_nulls(lhs, op, rhs, expected_text):
    p1 = Predicate(lhs, op, rhs)
    text = p1.render('?')
    assert text == expected_text
    assert len(p1.values) == 0


@pytest.mark.parametrize(
    'lhs, op, rhs', [
        ('foo', 'asdf', 99),
        ('bar', 6, 99),
        ('baz', 'UNLIKE', '%blah%'),
        ('baz', '<', None),
    ]
)
def test_bad_comp_op(lhs, op, rhs):
    with pytest.raises(InvalidComparisonOperator):
        _ = Predicate(lhs, op, rhs)


def test_query_1():
    query = Query()
    query.append(Table('my_table'))
    qtext, values = query.render('%s')
    assert qtext == 'SELECT * FROM "my_table"'

    query.append(Projection(['foo', 'bar', 'baz']))
    qtext, values = query.render('%s')
    assert qtext == 'SELECT "foo", "bar", "baz" FROM "my_table"'

    p1 = Predicate('foo', '!=', 0)
    p2 = Predicate('bar', 'LIKE', r'%blah%')
    where = Filter([p1, p2])
    query.append(where)
    qtext, values = query.render('%s')
    assert qtext == r'SELECT "foo", "bar", "baz" FROM "my_table" WHERE ("foo" != %s) AND ("bar" LIKE %s)'
    assert len(values) == 2
    assert values[0] == 0
    assert values[1] == r'%blah%'

    order = Order([('bar', Order.ASC), ('baz', Order.DESC)])
    query.append(order)
    qtext, values = query.render('%s')
    assert qtext == 'SELECT "foo", "bar", "baz" FROM "my_table" WHERE ("foo" != %s) AND ("bar" LIKE %s) ORDER BY "bar" ASC, "baz" DESC'
    assert len(values) == 2
    assert values[0] == 0
    assert values[1] == r'%blah%'


def test_query_2():
    query = Query()
    query.append(Table('my_table'))
    qtext, values = query.render('%s')
    assert qtext == 'SELECT * FROM "my_table"'

    p1 = Predicate('foo', '!=', 0)
    p2 = Predicate('bar', 'LIKE', r'%blah%')
    filt = Filter([p1, p2], Filter.OR)
    query.append(filt)
    qtext, values = query.render('%s')
    assert qtext == r'SELECT * FROM "my_table" WHERE (("foo" != %s) OR ("bar" LIKE %s))'
    assert len(values) == 2
    assert values[0] == 0
    assert values[1] == r'%blah%'

    query.append(Limit(10))
    qtext, values = query.render('%s')
    assert qtext == r'SELECT * FROM "my_table" WHERE (("foo" != %s) OR ("bar" LIKE %s)) LIMIT 10'
    assert len(values) == 2
    assert values[0] == 0
    assert values[1] == r'%blah%'

    query.append(Offset(10))
    qtext, values = query.render('?')
    assert qtext == r'SELECT * FROM "my_table" WHERE (("foo" != ?) OR ("bar" LIKE ?)) LIMIT 10 OFFSET 10'
    assert len(values) == 2
    assert values[0] == 0
    assert values[1] == r'%blah%'


def test_query_3():
    query = Query()
    query.append(Table('my_table'))
    p1 = Predicate('foo', '!=', 'null')
    p2 = Predicate('bar', 'LIKE', r'%blah%')
    where = Filter([p1, p2])
    query.append(where)
    qtext, values = query.render('%s')
    assert qtext == r'SELECT * FROM "my_table" WHERE ("foo" IS NOT NULL) AND ("bar" LIKE %s)'
    assert len(values) == 1
    assert values[0] == r'%blah%'

    query.append(Projection(['foo', 'bar', 'baz']))
    qtext, values = query.render('%s')
    assert qtext == 'SELECT "foo", "bar", "baz" FROM "my_table" WHERE ("foo" IS NOT NULL) AND ("bar" LIKE %s)'
    assert len(values) == 1
    assert values[0] == r'%blah%'


def test_filter_list():
    query = Query()
    query.append(Table('my_table'))
    p1 = Predicate('foo[*]', '=', 'bar')
    where = Filter([p1])
    query.append(where)
    qtext, values = query.render('%s')
    assert qtext == r'SELECT * FROM "my_table" WHERE ("foo" LIKE %s)'
    assert len(values) == 1
    assert values[0] == r'%bar%'


def test_filter_list_not_like():
    query = Query()
    query.append(Table('my_table'))
    p1 = Predicate('foo[*]', '!=', 'bar')
    where = Filter([p1])
    query.append(where)
    qtext, values = query.render('%s')
    assert qtext == r'SELECT * FROM "my_table" WHERE ("foo" NOT LIKE %s)'
    assert len(values) == 1
    assert values[0] == r'%bar%'


def test_filter_in():
    query = Query()
    query.append(Table('my_table'))
    p1 = Predicate('foo', 'IN', [1, 2, 3])
    where = Filter([p1])
    query.append(where)
    qtext, values = query.render('%s')
    assert qtext == r'SELECT * FROM "my_table" WHERE ("foo" IN (%s, %s, %s))'
    assert len(values) == 3
    assert values[0] == 1
    assert values[1] == 2
    assert values[2] == 3


def test_double_filter():
    query = Query()
    query.append(Table('my_table'))
    query.append(Filter([Predicate('foo', '=', 0),
                         Predicate('bar', '=', 1)], op=Filter.OR))
    query.append(Filter([Predicate('baz', '!=', 2),
                         Predicate('buz', '!=', 3)]))
    qtext, values = query.render('%s')
    assert qtext == r'SELECT * FROM "my_table" WHERE (("foo" = %s) OR ("bar" = %s)) AND ("baz" != %s) AND ("buz" != %s)'
    assert len(values) == 4
    assert values[0] == 0
    assert values[1] == 1
    assert values[2] == 2
    assert values[3] == 3


def test_filter_in():
    query = Query()
    query.append(Table('my_table'))
    p1 = Predicate('foo', 'IN', [1, 2, 3])
    where = Filter([p1])
    query.append(where)
    qtext, values = query.render('%s')
    assert qtext == r'SELECT * FROM "my_table" WHERE ("foo" IN (%s, %s, %s))'
    assert len(values) == 3
    assert values[0] == 1
    assert values[1] == 2
    assert values[2] == 3


def test_group():
    query = Query()
    query.append(Table('my_table'))
    p1 = Predicate('foo', '!=', 'null')
    p2 = Predicate('bar', 'LIKE', r'%blah%')
    query.append(Filter([p1, p2]))
    query.append(Group(['baz']))
    query.append(Aggregation([('SUM', 'foo', 'TotalFoo')]))
    qtext, values = query.render('%s')
    assert qtext == r'SELECT "baz", SUM("foo") AS "TotalFoo" FROM "my_table" WHERE ("foo" IS NOT NULL) AND ("bar" LIKE %s) GROUP BY "baz"'
    assert len(values) == 1
    assert values[0] == r'%blah%'
    query.append(Filter([Predicate('TotalFoo', '>=', 10)]))
    qtext, values = query.render('%s')
    assert qtext == r'SELECT "baz", SUM("foo") AS "TotalFoo" FROM "my_table" WHERE ("foo" IS NOT NULL) AND ("bar" LIKE %s) GROUP BY "baz" HAVING ("TotalFoo" >= %s)'
    assert len(values) == 2
    assert values[0] == r'%blah%'
    assert values[1] == 10


def test_group_with_proj():
    query = Query()
    query.append(Table('my_table'))
    query.append(Projection(['foo', 'bar', 'baz']))
    query.append(Group(['baz']))
    with pytest.raises(InvalidQuery) as e:
        query.append(Aggregation([('SUM', 'foo', 'TotalFoo')]))


def test_agg_without_group():
    query = Query()
    query.append(Table('my_table'))
    query.append(Aggregation([('SUM', 'foo', 'TotalFoo')]))
    qtext, values = query.render('%s')
    assert qtext == 'SELECT SUM("foo") AS "TotalFoo" FROM "my_table"'


def test_agg_without_alias():
    query = Query()
    query.append(Table('my_table'))
    query.append(Aggregation([('SUM', 'foo')]))
    qtext, values = query.render('%s')
    assert qtext == 'SELECT SUM("foo") AS "sum" FROM "my_table"'


def test_count_1():
    query = Query()
    query.append(Table('my_table'))
    query.append(Count())
    qtext, values = query.render('%s')
    assert qtext == 'SELECT COUNT(*) AS "count" FROM "my_table"'


def test_count_2():
    query = Query()
    query.append(Table('my_table'))
    query.append(Filter([Predicate('foo', '=', 1)]))
    query.append(Count())
    qtext, values = query.render('?')
    assert qtext == 'SELECT COUNT(*) AS "count" FROM "my_table" WHERE ("foo" = ?)'
    assert len(values) == 1
    assert values[0] == 1


def test_count_group():
    query = Query()
    query.append(Table('my_table'))
    query.append(Group(['baz']))
    query.append(Aggregation([('COUNT', None, 'count')]))
    qtext, values = query.render('?')
    # Since we group, we need the group value and count
    assert qtext == 'SELECT "baz", COUNT(*) AS "count" FROM "my_table" GROUP BY "baz"'


def test_unique():
    query = Query()
    query.append(Table('my_table'))
    query.append(Unique())
    qtext, values = query.render('%s')
    assert qtext == 'SELECT DISTINCT * FROM "my_table"'


def test_unique_count():
    query = Query()
    query.append(Table('my_table'))
    query.append(Unique())
    query.append(Count())
    qtext, values = query.render('%s')
    assert qtext == 'SELECT COUNT(*) AS "count" FROM (SELECT DISTINCT * FROM "my_table") AS tmp'


def test_proj_unique():
    query = Query()
    query.append(Table('my_table'))
    query.append(Projection(['foo', 'bar']))
    query.append(Unique())
    qtext, values = query.render('%s')
    assert qtext == 'SELECT DISTINCT "foo", "bar" FROM "my_table"'


def test_countunique():
    query = Query()
    query.append(Table('my_table'))
    query.append(CountUnique())
    qtext, values = query.render('%s')
    assert qtext == 'SELECT COUNT(*) AS "count" FROM (SELECT DISTINCT * FROM "my_table") AS tmp'


def test_proj_countunique():
    query = Query()
    query.append(Table('my_table'))
    query.append(Projection(['foo', 'bar']))
    query.append(CountUnique())
    qtext, values = query.render('%s')
    assert qtext == 'SELECT COUNT(DISTINCT "foo", "bar") AS "count" FROM "my_table"'


def test_proj_1_unique():
    query = Query()
    query.append(Table('my_table'))
    query.append(Projection(['foo']))
    query.append(Unique())
    qtext, values = query.render('%s')
    assert qtext == 'SELECT DISTINCT "foo" FROM "my_table"'


def test_proj_unique_count():
    query = Query()
    query.append(Table('my_table'))
    query.append(Projection(['foo', 'bar']))
    query.append(Unique())
    query.append(Count())
    qtext, values = query.render('%s')
    assert qtext == 'SELECT COUNT(DISTINCT "foo", "bar") AS "count" FROM "my_table"'


def test_join():
    query = Query()
    query.append(Table('left_table'))
    query.append(Join('right_table', 'left_col', '=', 'right_col'))
    qtext, values = query.render('%s')
    assert qtext == 'SELECT * FROM "left_table" INNER JOIN "right_table" ON "left_table"."left_col" = "right_table"."right_col"'


def test_join_3():
    query = Query()
    query.append(Table('table1'))
    query.append(Join('table2', 'col1', '=', 'col2'))
    query.append(Join('table3', 'col3', '=', 'col4'))
    qtext, values = query.render('%s')
    assert qtext == 'SELECT * FROM "table1" INNER JOIN "table2" ON "table1"."col1" = "table2"."col2" INNER JOIN "table3" ON "table2"."col3" = "table3"."col4"'


def test_join_filter():
    query = Query()
    query.append(Table('left_table'))
    query.append(Join('right_table', 'left_col', '=', 'right_col'))
    query.append(Filter([Predicate('foo', '=', 'bar')]))
    query.append(Projection(['baz']))
    query.append(Unique())
    qtext, values = query.render('%s')
    assert qtext == 'SELECT DISTINCT "baz" FROM "left_table" INNER JOIN "right_table" ON "left_table"."left_col" = "right_table"."right_col" WHERE ("foo" = %s)'
    assert len(values) == 1
    assert values[0] == 'bar'


def test_join_without_table():
    query = Query()
    with pytest.raises(InvalidQuery):
        query.append(Join('right_table', 'left_col', '=', 'right_col'))


def test_implicit_table():
    query = Query('my_table')
    qtext, values = query.render('%s')
    assert qtext == 'SELECT * FROM "my_table"'

    query.append(Projection(['foo', 'bar', 'baz']))
    query.append(Order(['foo']))
    qtext, values = query.render('%s')
    assert qtext == 'SELECT "foo", "bar", "baz" FROM "my_table" ORDER BY "foo" ASC'


def test_query_init_list():
    query = Query([Table('my_table'), Projection(['foo', 'bar', 'baz'])])
    qtext, values = query.render('%s')
    assert qtext == 'SELECT "foo", "bar", "baz" FROM "my_table"'


def test_init_list_join_filter():
    # Same as test_join_filter but using init list
    query = Query([
        Table('left_table'),
        Join('right_table', 'left_col', '=', 'right_col'),
        Filter([Predicate('foo', '=', 'bar')]),
        Projection(['baz']),
        Unique(),
    ])
    qtext, values = query.render('%s')
    assert qtext == 'SELECT DISTINCT "baz" FROM "left_table" INNER JOIN "right_table" ON "left_table"."left_col" = "right_table"."right_col" WHERE ("foo" = %s)'
    assert len(values) == 1
    assert values[0] == 'bar'



import ipaddress
import logging
import os
import re
import sqlite3

from firepit.exceptions import DuplicateTable
from firepit.exceptions import InvalidAttr
from firepit.exceptions import UnexpectedError
from firepit.exceptions import UnknownViewname
from firepit.splitter import SqlWriter
from firepit.sqlstorage import SqlStorage
from firepit.sqlstorage import infer_type
from firepit.sqlstorage import validate_name

logger = logging.getLogger(__name__)


def get_storage(path):
    return SQLiteStorage(path)


def _in_subnet(value, net):
    """User-defined function to help implement STIX ISSUBSET"""
    if '/' in value:
        value = ipaddress.IPv4Network(value).network_address
    else:
        value = ipaddress.IPv4Address(value)
    net = ipaddress.IPv4Network(net)
    return value in net


def _match(pattern, value):
    """User-defined function to implement SQL MATCH/STIX MATCHES"""
    return bool(re.match(pattern, value))


class SQLiteStorage(SqlStorage):
    def __init__(self, dbname):
        super().__init__()
        self.placeholder = '?'
        self.dbname = dbname
        self.connection = sqlite3.connect(dbname)
        self.connection.row_factory = row_factory
        logger.debug("Connection to SQLite DB %s successful", dbname)

        # Create functions for IP address subnet membership checks
        self.connection.create_function('in_subnet', 2, _in_subnet)

        # Create function for SQL MATCH
        self.connection.create_function("match", 2, _match)

        # Do DB initization
        cursor = self.connection.cursor()
        cursor.execute('BEGIN;')
        self._initdb(cursor)
        cursor.close()

    def _get_writer(self, **kwargs):
        """Get a DB inserter object"""
        filedir = os.path.dirname(self.dbname)
        return SqlWriter(filedir, self, infer_type=infer_type)

    def _do_execute(self, query, values=None, cursor=None):
        if not cursor:
            cursor = self.connection.cursor()
        try:
            logger.debug('Executing query: %s', query)
            if not values:
                cursor.execute(query)
            else:
                cursor.execute(query, values)
        except sqlite3.OperationalError as e:
            logger.debug('%s', e)  #, exc_info=e)
            if e.args[0].startswith("no such column"):
                m = e.args[0].replace("no such column", "invalid attribute")
                raise InvalidAttr(m) from e
            elif e.args[0].startswith("no such table: main."):
                # Just means no match - return empty cursor?
                cursor = self.connection.cursor()
            elif e.args[0].startswith("no such table: "):
                raise UnknownViewname(e.args[0]) from e
            elif e.args[0].endswith("syntax error"):
                # We see this on SQL injection attempts
                raise UnexpectedError(e.args[0]) from e
            else:
                raise e  # See if caller wants special behavior
        return cursor

    def _execute(self, statement, cursor=None):
        return self._do_execute(statement, cursor=cursor)

    def _query(self, query, values=None, cursor=None):
        cursor = self._do_execute(query, values=values, cursor=cursor)
        self.connection.commit()
        return cursor

    def _create_view(self, viewname, select, sco_type, deps=None, cursor=None):
        """Overrides parent"""
        validate_name(viewname)
        if not cursor:
            cursor = self._execute('BEGIN;')
        is_new = True
        if not deps:
            deps = []
        elif viewname in deps:
            is_new = False
            # Get the query that makes up the current view
            slct = self._get_view_def(viewname)
            if self._is_sql_view(viewname, cursor):
                self._execute(f'DROP VIEW IF EXISTS "{viewname}"', cursor)
            else:
                self._execute(f'ALTER TABLE "{viewname}" RENAME TO "_{viewname}"', cursor)
                slct = slct.replace(viewname, f'_{viewname}')
            # Swap out the viewname for its definition
            select = re.sub(f'"{viewname}"', f'({slct}) AS tmp', select)
        if self._is_sql_view(viewname, cursor):
            is_new = False
            self._execute(f'DROP VIEW IF EXISTS "{viewname}"', cursor)
        self._execute(f'CREATE VIEW "{viewname}" AS {select}', cursor)
        if is_new:
            self._new_name(cursor, viewname, sco_type)
        return cursor

    def _create_table(self, tablename, columns):
        stmt = f'CREATE TABLE "{tablename}" ('
        stmt += ','.join([f'"{colname}" {coltype}' for colname, coltype in columns.items()])
        stmt += ');'
        logger.debug('_create_table: "%s"', stmt)
        try:
            cursor = self._execute(stmt)
        except sqlite3.OperationalError as e:
            self.connection.rollback()
            logger.debug('_create_table: %s', e)  #, exc_info=e)
            if e.args[0].startswith(f'table "{tablename}" already exists'):
                raise DuplicateTable(tablename) from e
        if 'x_contained_by_ref' in columns:
            self._execute(f'CREATE INDEX "{tablename}_obs" ON "{tablename}" ("x_contained_by_ref");', cursor)
        self.connection.commit()
        cursor.close()

    def _add_column(self, tablename, prop_name, prop_type):
        stmt = f'ALTER TABLE "{tablename}" ADD COLUMN "{prop_name}" {prop_type};'
        logger.debug('new_property: "%s"', stmt)
        try:
            cursor = self._execute(stmt)
            self.connection.commit()
            cursor.close()
        except sqlite3.OperationalError as e:
            self.connection.rollback()
            logger.debug('%s', e)  #, exc_info=e)
            if e.args[0].startswith('duplicate column name: '):
                pass
            else:
                raise Exception('Internal error: ' + e.args[0]) from e

    def _get_view_def(self, viewname):
        view = self._query(("SELECT sql from sqlite_master"
                            " WHERE type='view' and name=?"),
                           values=(viewname,)).fetchone()
        if view:
            slct = view['sql']
            return slct.replace(f'CREATE VIEW "{viewname}" AS ', '')

        # Must be a table
        return f'SELECT * FROM "{viewname}"'

    def _is_sql_view(self, name, cursor=None):
        view = self._query(("SELECT sql from sqlite_master"
                            " WHERE type='view' and name=?"),
                           values=(name,)).fetchone()
        return view is not None

    def tables(self):
        cursor = self.connection.execute(
            "SELECT name FROM sqlite_master WHERE type='table';")
        rows = cursor.fetchall()
        return [i['name'] for i in rows
                if not i['name'].startswith('__') and
                not i['name'].startswith('sqlite')]

    def types(self):
        stmt = ("SELECT name FROM sqlite_master WHERE type='table'"
                " EXCEPT SELECT name FROM __symtable")
        cursor = self.connection.execute(stmt)
        rows = cursor.fetchall()
        return [i['name'] for i in rows
                if not i['name'].startswith('__') and
                not i['name'].startswith('sqlite')]

    def columns(self, viewname):
        validate_name(viewname)
        stmt = f'PRAGMA table_info("{viewname}")'
        cursor = self._execute(stmt)
        try:
            mappings = cursor.fetchall()
            if mappings:
                result = [e["name"] for e in mappings]
            else:
                result = []
            logger.debug('%s columns = %s', viewname, result)
        except sqlite3.OperationalError as e:
            logger.error('%s', e)
            result = []
        return result

    def schema(self, viewname):
        validate_name(viewname)
        stmt = f'PRAGMA table_info("{viewname}")'
        cursor = self._execute(stmt)
        return [{k: v for k, v in row.items() if k in ['name', 'type']}
                for row in cursor.fetchall()]

    def delete(self):
        """Delete ALL data in this store"""
        self.connection.close()
        try:
            os.remove(self.dbname)
        except FileNotFoundError:
            pass


def row_factory(cursor, row):
    return {col[0]: row[idx] for idx, col in enumerate(cursor.description)}



import os
from collections import defaultdict
from lark import Lark, Transformer, v_args


def get_grammar():
    pth = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                       "paramstix.lark")
    return open(pth, "r").read()


def stix2sql(pattern, sco_type):
    grammar = get_grammar()
    return Lark(grammar,
                parser="lalr",
                transformer=_TranslateTree(sco_type)).parse(pattern)


@v_args(inline=True)
class _TranslateTree(Transformer):
    """Transformer to convert relevant parts of STIX pattern to WHERE clause"""

    def __init__(self, sco_type):
        self.sco_type = sco_type

    def _make_comp(self, lhs, op, rhs):
        orig_op = op
        sco_type, _, prop = lhs.partition(':')

        # Ignore object paths that don't match table type
        if self.sco_type == sco_type:
            neg, _, op = op.rpartition(' ')
            if op == 'ISSUBSET':
                op = '=='
                if sco_type == 'ipv4-addr' or lhs in ['network-traffic:src_ref.value',
                                                      'network-traffic:dst_ref.value']:
                    return f'{neg} (in_subnet("{prop}", {rhs}))'
                else:
                    raise ValueError(
                        f'{orig_op} not supported for SCO type {sco_type}')
            elif op == 'ISSUPERSET':  # When would anyone use ISSUPERSET?
                op = '=='
                if sco_type == 'ipv4-addr' or lhs in ['network-traffic:src_ref.value',
                                                      'network-traffic:dst_ref.value']:
                    return f'{neg} (in_subnet({rhs}, "{prop}"))'  # FIXME!
                else:
                    raise ValueError(
                        f'{orig_op} not supported for SCO type {sco_type}')
            elif 'MATCHES' in op:
                return f'{neg} match({rhs}, "{prop}")'
            prop, chunk, subprop = prop.partition('[*]')
            if chunk:
                if op == '!=':
                    neg = 'NOT'
                op = 'LIKE'
                rhs = rhs.strip("'")
                if subprop:
                    subprop = subprop.lstrip('.')
                    rhs = f"'%\"{subprop}\":\"{rhs}\"%'"
                else:
                    rhs = f"'%{rhs}%'"
            return f'"{prop}" {neg} {op} {rhs}'
        return ''

    def _make_exp(self, lhs, op, rhs):
        return op.join(filter(None, [lhs, rhs]))

    def disj(self, lhs, rhs):
        return self._make_exp(lhs, ' OR ', rhs)

    def conj(self, lhs, rhs):
        return self._make_exp(lhs, ' AND ', rhs)

    def obs_disj(self, lhs, rhs):
        return self.disj(lhs, rhs)

    def obs_conj(self, lhs, rhs):
        return self.conj(lhs, rhs)

    def comp_grp(self, exp):
        return f'({exp})'

    def simple_comp_exp(self, lhs, op, rhs):
        return self._make_comp(lhs, op, rhs)

    def comp_disj(self, lhs, rhs):
        return self.disj(lhs, rhs)

    def comp_conj(self, lhs, rhs):
        return self.conj(lhs, rhs)

    def op(self, value):
        return f'{value}'

    def quoted_str(self, value):
        return f"'{value}'"

    def lit_list(self, *args):
        return "(" + ','.join(args) + ")"

    def start(self, exp, qualifier):
        # For now, drop the qualifier.  Assume the query handled it.
        return f'{exp}'

    def object_path(self, sco_type, prop):
        return f'{sco_type}:{prop}'


def summarize_pattern(pattern):
    grammar = get_grammar()
    paths = Lark(grammar,
                 parser="lalr",
                 transformer=_SummarizePattern()).parse(pattern)
    result = defaultdict(set)
    for path in paths:
        sco_type, _, prop = path.partition(':')
        result[sco_type].add(prop)
    return result


@v_args(inline=True)
class _SummarizePattern(Transformer):
    def obs_disj(self, lhs, rhs):
        return lhs | rhs

    def obs_conj(self, lhs, rhs):
        return lhs & rhs

    def comp_grp(self, exp):
        return exp

    def simple_comp_exp(self, lhs, _op, _rhs):
        return {lhs}

    def comp_disj(self, lhs, rhs):
        return lhs | rhs

    def comp_conj(self, lhs, rhs):
        return lhs | rhs  # Still want union here

    # None of these actually matter
    def op(self, _op):
        return None

    def quoted_str(self, _value):
        return None

    def lit_list(self, *args):
        return None

    def start(self, exp, _qualifier):
        return exp

    def object_path(self, sco_type, prop):
        return f'{sco_type}:{prop}'



"""Top-level package for STIX Columnar Storage."""

__author__ = """IBM Security"""
__email__ = 'pcoccoli@us.ibm.com'
__version__ = '1.3.6'


from importlib import import_module
from urllib.parse import urlparse

from firepit.validate import validate_name


def get_storage(url, session_id=None):
    if session_id:
        validate_name(session_id)
    url = urlparse(url)
    if url.scheme == 'postgresql':
        module = import_module('firepit.pgstorage')
        return module.get_storage(url, session_id)
    if url.scheme in ['sqlite3', '']:
        module = import_module('firepit.sqlitestorage')
        return module.get_storage(url.path)
    raise NotImplementedError(url.scheme)



"""STIX and SQL identifier validators"""

import re

from firepit.exceptions import InvalidStixPath
from firepit.exceptions import InvalidViewname


def validate_name(name):
    if not bool(re.match(r'^[\w-]*$', name)):
        raise InvalidViewname(name)


def validate_path(path):
    if not bool(re.match(r"^([a-zA-Z][a-zA-Z0-9-]*:)?[\w\'-]+(\[\*\])?(\.[\w\'-]+)*$", path)):
        raise InvalidStixPath(path)
